"""Define the scene module."""
from typing import Dict, Optional

from .api import Api
from .entity import Entity


class Scene:
    """Define a scene data entity."""

    def __init__(self):
        """Create a new instance of the Scene class."""
        self._color = None
        self._icon = None
        self._location_id = None
        self._name = None
        self._scene_id = None

    def apply_data(self, data: dict):
        """Apply the data structure to the class."""
        self._color = data['sceneColor']
        self._icon = data['sceneIcon']
        self._location_id = data['locationId']
        self._name = data['sceneName']
        self._scene_id = data['sceneId']

    @property
    def color(self) -> str:
        """Get the color of the scene."""
        return self._color

    @property
    def icon(self) -> str:
        """Get the icon of the scene."""
        return self._icon

    @property
    def location_id(self) -> str:
        """Get the location this scene is in."""
        return self._location_id

    @property
    def name(self) -> str:
        """Get the name of the scene."""
        return self._name

    @property
    def scene_id(self) -> str:
        """Get the id of the scene."""
        return self._scene_id


class SceneEntity(Entity, Scene):
    """Define a scene entity."""

    def __init__(self, api: Api, data: Optional[Dict] = None):
        """Create a new instance of the class."""
        Entity.__init__(self, api)
        Scene.__init__(self)
        if data:
            self.apply_data(data)

    async def execute(self):
        """Execute the scene."""
        result = await self._api.execute_scene(self._scene_id)
        return result == {'status': 'success'}

    async def refresh(self):
        """Refresh is not implemented."""
        raise NotImplementedError()

    async def save(self):
        """Save is not implemented."""
        raise NotImplementedError()



"""Tests for the room module."""
import pytest

from pysmartthings.room import Room, RoomEntity

from .conftest import LOCATION_ID, ROOM_ID
from .utilities import get_json


class TestRoom:
    """Tests for the Room class."""

    @staticmethod
    def test_apply_data():
        """Test the init method."""
        # Arrange
        data = get_json('room.json')
        room = Room()
        room.apply_data(data)
        # Assert
        assert room.room_id == ROOM_ID
        assert room.location_id == LOCATION_ID
        assert room.name == 'Theater'
        assert room.background_image == 'Test'


class TestRoomEntity:
    """Tests for the room entity class."""

    @staticmethod
    @pytest.mark.asyncio
    async def test_refresh(api):
        """Tests the refresh method."""
        # Arrange
        entity = RoomEntity(api, location_id=LOCATION_ID, room_id=ROOM_ID)
        # Act
        await entity.refresh()
        # Assert
        assert entity.name == 'Theater'

    @staticmethod
    @pytest.mark.asyncio
    async def test_save(api):
        """Tests the save method."""
        # Arrange
        entity = RoomEntity(api, location_id=LOCATION_ID, room_id=ROOM_ID)
        entity.name = 'Theater'
        entity.background_image = 'Test'
        # Act
        await entity.save()
        # Assert
        assert entity.name == 'Theater'
        assert entity.background_image == 'Test'



"""Define the installedapp module."""

from enum import Enum
from typing import Sequence

from .api import Api
from .entity import Entity
from .subscription import SubscriptionEntity


class InstalledAppType(Enum):
    """Define the type of installed app."""

    UNKNOWN = "UNKNOWN"
    LAMBDA_SMART_APP = 'LAMBDA_SMART_APP'
    WEBHOOK_SMART_APP = 'WEBHOOK_SMART_APP'
    BEHAVIOR = 'BEHAVIOR'


class InstalledAppStatus(Enum):
    """Define the installed app status."""

    UNKNOWN = "UNKNOWN"
    PENDING = "PENDING"
    AUTHORIZED = "AUTHORIZED"
    REVOKED = "REVOKED"
    DISABLED = "DISABLED"


class InstalledApp:
    """Define the InstalledApp class."""

    def __init__(self):
        """Create a new instance of the InstalledApp class."""
        self._installed_app_id = None
        self._installed_app_type = InstalledAppType.UNKNOWN
        self._installed_app_status = InstalledAppStatus.UNKNOWN
        self._display_name = None
        self._app_id = None
        self._reference_id = None
        self._location_id = None
        self._created_date = None
        self._last_updated_date = None
        self._classifications = []

    def apply_data(self, data: dict):
        """Apply the data structure to the properties."""
        self._installed_app_id = data['installedAppId']
        self._installed_app_type = InstalledAppType(data['installedAppType'])
        self._installed_app_status = \
            InstalledAppStatus(data['installedAppStatus'])
        self._display_name = data['displayName']
        self._app_id = data['appId']
        self._reference_id = data['referenceId']
        self._location_id = data['locationId']
        self._created_date = data['createdDate']
        self._last_updated_date = data['lastUpdatedDate']
        self._classifications = data['classifications']

    @property
    def installed_app_id(self) -> str:
        """Get the ID of the installed app."""
        return self._installed_app_id

    @property
    def installed_app_type(self) -> InstalledAppType:
        """Get the type of installed app."""
        return self._installed_app_type

    @property
    def installed_app_status(self) -> InstalledAppStatus:
        """Get the current state of an install."""
        return self._installed_app_status

    @property
    def display_name(self) -> str:
        """Get the user defined name for the installed app."""
        return self._display_name

    @property
    def app_id(self) -> str:
        """Get the ID of the app."""
        return self._app_id

    @property
    def reference_id(self) -> str:
        """Get a reference to an upstream system."""
        return self._reference_id

    @property
    def location_id(self) -> str:
        """Get the ID of the location to which the installed app."""
        return self._location_id

    @property
    def created_date(self) -> str:
        """Get the date the installed app was created."""
        return self._created_date

    @property
    def last_updated_date(self) -> str:
        """Get the date the installed app was updated."""
        return self._last_updated_date

    @property
    def classifications(self) -> Sequence[str]:
        """Get the collection of classifications."""
        return self._classifications


class InstalledAppEntity(Entity, InstalledApp):
    """Define the InstalledAppEntity class."""

    def __init__(self, api: Api, data=None, installed_app_id=None):
        """Create a new instance of the InstalledAppEntity class."""
        Entity.__init__(self, api)
        InstalledApp.__init__(self)
        if data:
            self.apply_data(data)
        if installed_app_id:
            self._installed_app_id = installed_app_id

    async def refresh(self):
        """Refresh the installedapp information using the API."""
        data = await self._api.get_installed_app(self._installed_app_id)
        self.apply_data(data)

    async def save(self):
        """Save the changes made to the app."""
        raise NotImplementedError

    async def subscriptions(self) -> Sequence[SubscriptionEntity]:
        """Get the subscriptions for the installedapp."""
        data = await self._api.get_subscriptions(self._installed_app_id)
        return [SubscriptionEntity(self._api, entity) for entity in data]



"""Define the SmartThings Cloud API."""

from typing import List, Optional, Sequence

from aiohttp import ClientSession

from .api import Api
from .app import (
    App, AppEntity, AppOAuth, AppOAuthClient, AppOAuthClientEntity,
    AppOAuthEntity, AppSettings, AppSettingsEntity)
from .device import DeviceEntity
from .installedapp import InstalledAppEntity, InstalledAppStatus
from .location import LocationEntity
from .oauthtoken import OAuthToken
from .room import Room, RoomEntity
from .scene import SceneEntity
from .subscription import Subscription, SubscriptionEntity


class SmartThings:
    """Define a class for interacting with the SmartThings Cloud API."""

    __slots__ = ['_service']

    def __init__(self, session: ClientSession, token: str):
        """Initialize the SmartThingsApi."""
        self._service = Api(session, token)

    async def locations(self) -> List[LocationEntity]:
        """Retrieve SmartThings locations."""
        resp = await self._service.get_locations()
        return [LocationEntity(self._service, entity) for entity in resp]

    async def location(self, location_id: str) -> LocationEntity:
        """Retrieve a location with the specified ID."""
        entity = await self._service.get_location(location_id)
        return LocationEntity(self._service, entity)

    async def rooms(self, location_id: str) -> List[RoomEntity]:
        """Retrieve a list of rooms for a location."""
        resp = await self._service.get_rooms(location_id)
        return [RoomEntity(self._service, entity) for entity in resp]

    async def room(self, location_id: str, room_id: str) -> RoomEntity:
        """Retrieve a specific room."""
        entity = await self._service.get_room(location_id, room_id)
        return RoomEntity(self._service, entity)

    async def create_room(self, room: Room) -> RoomEntity:
        """Create a room."""
        entity = await self._service.create_room(
            room.location_id, room.to_data())
        return RoomEntity(self._service, entity)

    async def update_room(self, room: Room) -> RoomEntity:
        """Update a room."""
        entity = await self._service.update_room(
            room.location_id, room.room_id, room.to_data())
        return RoomEntity(self._service, entity)

    async def delete_room(self, location_id: str, room_id: str):
        """Delete a room."""
        return await self._service.delete_room(location_id, room_id) == {}

    async def devices(self, *, location_ids: Optional[Sequence[str]] = None,
                      capabilities: Optional[Sequence[str]] = None,
                      device_ids: Optional[Sequence[str]] = None) -> List:
        """Retrieve SmartThings devices."""
        params = []
        if location_ids:
            params.extend([('locationId', lid) for lid in location_ids])
        if capabilities:
            params.extend([('capability', cap) for cap in capabilities])
        if device_ids:
            params.extend([('deviceId', did) for did in device_ids])
        resp = await self._service.get_devices(params)
        return [DeviceEntity(self._service, entity) for entity in resp]

    async def device(self, device_id: str) -> DeviceEntity:
        """Retrieve a device with the specified ID."""
        entity = await self._service.get_device(device_id)
        return DeviceEntity(self._service, entity)

    async def apps(self, *, app_type: Optional[str] = None) -> List[AppEntity]:
        """Retrieve list of apps."""
        params = []
        if app_type:
            params.append(('appType', app_type))
        resp = await self._service.get_apps(params)
        return [AppEntity(self._service, entity) for entity in resp]

    async def app(self, app_id: str) -> AppEntity:
        """Retrieve an app with the specified ID."""
        entity = await self._service.get_app(app_id)
        return AppEntity(self._service, entity)

    async def create_app(self, app: App) -> (AppEntity, AppOAuthClient):
        """Create a new app."""
        entity = await self._service.create_app(app.to_data())
        return AppEntity(self._service, entity['app']), AppOAuthClient(entity)

    async def delete_app(self, app_id: str):
        """Delete an app."""
        return await self._service.delete_app(app_id) == {}

    async def app_settings(self, app_id: str) -> AppSettingsEntity:
        """Get an app's settings."""
        settings = await self._service.get_app_settings(app_id)
        return AppSettingsEntity(self._service, app_id, settings)

    async def update_app_settings(self, data: AppSettings) -> \
            AppSettingsEntity:
        """Update an app's settings."""
        entity = await self._service.update_app_settings(
            data.app_id, data.to_data())
        return AppSettingsEntity(self._service, data.app_id, entity)

    async def app_oauth(self, app_id: str) -> AppOAuthEntity:
        """Get an app's OAuth settings."""
        oauth = await self._service.get_app_oauth(app_id)
        return AppOAuthEntity(self._service, app_id, oauth)

    async def update_app_oauth(self, data: AppOAuth) -> AppOAuthEntity:
        """Update an app's OAuth settings without having to retrieve it."""
        entity = await self._service.update_app_oauth(
            data.app_id, data.to_data())
        return AppOAuthEntity(self._service, data.app_id, entity)

    async def generate_app_oauth(
            self, data: AppOAuth) -> AppOAuthClientEntity:
        """Generate a new oauth client id and secret."""
        entity = await self._service.generate_app_oauth(
            data.app_id, data.to_data())
        return AppOAuthClientEntity(self._service, data.app_id, entity)

    async def installed_apps(
            self, *, location_id: Optional[str] = None,
            installed_app_status: Optional[InstalledAppStatus] = None) -> \
            List[InstalledAppEntity]:
        """Get a list of the installed applications."""
        params = []
        if location_id:
            params.append(('locationId', location_id))
        if installed_app_status:
            params.append(('installedAppStatus', installed_app_status.value))
        resp = await self._service.get_installed_apps(params)
        return [InstalledAppEntity(self._service, entity) for entity in resp]

    async def installed_app(self, installed_app_id: str) -> InstalledAppEntity:
        """Get an installedapp with the specified ID."""
        entity = await self._service.get_installed_app(installed_app_id)
        return InstalledAppEntity(self._service, entity)

    async def delete_installed_app(self, installed_app_id: str):
        """Delete an installedapp."""
        result = await self._service.delete_installed_app(installed_app_id)
        return result == {'count': 1}

    async def subscriptions(self, installed_app_id: str) \
            -> List[SubscriptionEntity]:
        """Get an installedapp's subscriptions."""
        resp = await self._service.get_subscriptions(installed_app_id)
        return [SubscriptionEntity(self._service, entity)
                for entity in resp]

    async def delete_subscriptions(self, installed_app_id: str) -> int:
        """Delete an installedapp's subscriptions."""
        resp = await self._service.delete_all_subscriptions(installed_app_id)
        return resp['count']

    async def delete_subscription(self, installed_app_id: str,
                                  subscription_id: str):
        """Delete an individual subscription."""
        return await self._service.delete_subscription(
            installed_app_id, subscription_id) == {'count': 1}

    async def create_subscription(self, subscription: Subscription) \
            -> SubscriptionEntity:
        """Create a new subscription for an installedapp."""
        entity = await self._service.create_subscription(
            subscription.installed_app_id, subscription.to_data())
        return SubscriptionEntity(self._service, entity)

    async def scenes(self, *, location_id: Optional[str] = None):
        """Get a list of scenes and optionally filter by location."""
        params = []
        if location_id:
            params.append(('locationId', location_id))
        resp = await self._service.get_scenes(params)
        return [SceneEntity(self._service, entity) for entity in resp]

    async def execute_scene(self, scene_id: str) -> bool:
        """Execute the scene with the specified id."""
        result = await self._service.execute_scene(scene_id)
        return result == {'status': 'success'}

    async def generate_tokens(
            self, client_id: str, client_secret: str,
            refresh_token: str) -> OAuthToken:
        """Generate a new refresh/access token pair."""
        result = await self._service.generate_tokens(
            client_id, client_secret, refresh_token)
        return OAuthToken(self._service, result)



"""Define the SmartThing location."""

from typing import List, Optional

from .api import Api
from .entity import Entity
from .room import RoomEntity


class Location:
    """Represents a SmartThings Location."""

    def __init__(self):
        """Initialize a new location."""
        self._name = None
        self._location_id = None
        self._latitude = None
        self._longitude = None
        self._region_radius = None
        self._temperature_scale = None
        self._locale = None
        self._country_code = None
        self._timezone_id = None

    def apply_data(self, data: dict):
        """Apply the given data structure to the location."""
        self._name = data['name']
        self._location_id = data['locationId']
        self._latitude = data.get('latitude', None)
        self._longitude = data.get('longitude', None)
        self._region_radius = data.get('regionRadius', None)
        self._temperature_scale = data.get('temperatureScale', None)
        self._locale = data.get('locale', None)
        self._country_code = data.get('countryCode', None)
        self._timezone_id = data.get('timeZoneId', None)

    @property
    def name(self) -> str:
        """Get nickname given for the location."""
        return self._name

    @property
    def location_id(self) -> str:
        """Get the ID of the location."""
        return self._location_id

    @property
    def latitude(self) -> float:
        """Get the geographical latitude."""
        return self._latitude

    @property
    def longitude(self) -> float:
        """Get the geographical longitude."""
        return self._longitude

    @property
    def region_radius(self) -> int:
        """Get the radius in meters which defines this location."""
        return self._region_radius

    @property
    def temperature_scale(self) -> str:
        """Get the temperature scale of the location (F or C)."""
        return self._temperature_scale

    @property
    def locale(self) -> str:
        """Get the IETF BCP 47 language tag of the location."""
        return self._locale

    @property
    def country_code(self) -> str:
        """Get the country code of the location."""
        return self._country_code

    @property
    def timezone_id(self) -> str:
        """Get the ID matching the Java Time Zone ID of the location."""
        return self._timezone_id


class LocationEntity(Entity, Location):
    """Define a location entity."""

    def __init__(self, api: Api, data: Optional[dict] = None,
                 location_id: Optional[str] = None):
        """Create a new instance of the LocationEntity."""
        Entity.__init__(self, api)
        Location.__init__(self)
        if data:
            self.apply_data(data)
        if location_id:
            self._location_id = location_id

    async def refresh(self):
        """Refresh the location information."""
        data = await self._api.get_location(self._location_id)
        if data:
            self.apply_data(data)

    async def save(self):
        """Location does not support updating at this time."""
        raise NotImplementedError(
            'Location does not support updating at this time.')

    async def rooms(self) -> List[RoomEntity]:
        """Get the rooms contained within the location."""
        resp = await self._api.get_rooms(self._location_id)
        return [RoomEntity(self._api, entity) for entity in resp]



"""Define test configuration."""
import glob
import re

import pytest

from pysmartthings.api import (
    API_APP, API_APP_OAUTH, API_APP_OAUTH_GENERATE, API_APP_SETTINGS, API_APPS,
    API_BASE, API_DEVICE, API_DEVICE_COMMAND, API_DEVICE_STATUS, API_DEVICES,
    API_INSTALLEDAPP, API_INSTALLEDAPPS, API_LOCATION, API_LOCATIONS,
    API_OAUTH_TOKEN, API_ROOM, API_ROOMS, API_SCENE_EXECUTE, API_SCENES,
    API_SUBSCRIPTION, API_SUBSCRIPTIONS, Api)
from pysmartthings.smartthings import SmartThings

from .utilities import ClientMocker

APP_ID = 'c6cde2b0-203e-44cf-a510-3b3ed4706996'
AUTH_TOKEN = '9b3fd445-42d6-441b-b386-99ea51e13cb0'
CLIENT_ID = '7cd4d474-7b36-4e03-bbdb-4cd4ae45a2be'
CLIENT_SECRET = '9b3fd445-42d6-441b-b386-99ea51e13cb0'
DEVICE_ID = '743de49f-036f-4e9c-839a-2f89d57607db'
INSTALLED_APP_ID = '4514eb36-f5fd-4ab2-9520-0597acd1d212'
LOCATION_ID = '397678e5-9995-4a39-9d9f-ae6ba310236b'
ROOM_ID = '7715151d-0314-457a-a82c-5ce48900e065'
REFRESH_TOKEN = 'a86a5c8e-0014-44a6-8980-5846633972dd'
SUBSCRIPTION_ID = '7bdf5909-57c4-41f3-9089-e520513bd92a'
SCENE_ID = '9b58411f-5d26-418d-b193-3434a77c484a'

DEVICE_COMMAND_PATTERN = re.compile(r'(device_command_post_[a-z_]+)')


def register_device_commands(mocker):
    """Register all device commands."""
    files = glob.glob('tests/json/device_command_post_*.json')
    for file in files:
        match = DEVICE_COMMAND_PATTERN.search(file)
        if match:
            mocker.post(API_DEVICE_COMMAND.format(device_id=DEVICE_ID),
                        request=match.group(), response={})


def register_url_mocks(mocker):
    """Register the URLs we need to mock."""
    mocker.default_headers = {'Authorization': "Bearer " + AUTH_TOKEN}
    mocker.base_url = API_BASE

    # Locations
    mocker.get(API_LOCATIONS, response='locations')
    mocker.get(API_LOCATION.format(location_id=LOCATION_ID),
               response='location')

    # Rooms
    mocker.get(API_ROOMS.format(location_id=LOCATION_ID), response='rooms')
    mocker.get(API_ROOM.format(location_id=LOCATION_ID, room_id=ROOM_ID),
               response='room')
    mocker.post(API_ROOMS.format(location_id=LOCATION_ID),
                request='room_post', response='room')
    mocker.put(API_ROOM.format(location_id=LOCATION_ID, room_id=ROOM_ID),
               request='room_put', response='room')
    mocker.delete(API_ROOM.format(location_id=LOCATION_ID, room_id=ROOM_ID),
                  response={})

    # Devices
    mocker.get(API_DEVICES, response='devices')
    mocker.get(API_DEVICES, response='devices_filtered', params=[
        ('locationId', LOCATION_ID),
        ('capability', 'switch'),
        ('deviceId', 'edd26ac6-d156-4505-9647-3b20118ae4d1'),
        ('deviceId', 'be1a61ce-c2a4-4b32-bf8c-31de6d3fa7dd')])
    mocker.get(API_DEVICE.format(device_id=DEVICE_ID), response='device')
    mocker.get(API_DEVICE_STATUS.format(device_id=DEVICE_ID),
               response='device_status')

    # Device Commands
    register_device_commands(mocker)

    # Apps
    mocker.get(API_APPS, response='apps')
    mocker.get(API_APP.format(app_id=APP_ID), response='app_get')
    mocker.post(API_APPS, request='app_post_request',
                response='app_post_response')
    mocker.put(API_APP.format(app_id=APP_ID), request='app_put_request',
               response='app_put_response')
    mocker.delete(API_APP.format(app_id=APP_ID), response={})
    mocker.get(API_APP_SETTINGS.format(app_id=APP_ID),
               response='app_settings')
    mocker.put(API_APP_SETTINGS.format(app_id=APP_ID), request='app_settings',
               response='app_settings')
    mocker.get(API_APP_OAUTH.format(app_id=APP_ID),
               response='app_oauth_get_response')
    mocker.put(API_APP_OAUTH.format(app_id=APP_ID),
               request='app_oauth_put_request',
               response='app_oauth_put_response')
    mocker.post(API_APP_OAUTH_GENERATE.format(app_id=APP_ID),
                request='app_oauth_generate_request',
                response='app_oauth_generate_response')

    # InstalledApps
    mocker.get(API_INSTALLEDAPPS, response='installedapps_get_response')
    mocker.request('get', "https://api.smartthings.com/installedapps?"
                   "currentLocationId=NWMwM2U1MTgtMTE4YS00NGNiLTg1YWQtNzg3N2Qw"
                   "YjMwMmU0&currentOffset=MA", headers=mocker.default_headers,
                   response='installedapps_get_response_2')
    mocker.get(API_INSTALLEDAPP.format(installed_app_id=INSTALLED_APP_ID),
               response='installedapp_get_response')
    mocker.delete(
        API_INSTALLEDAPP.format(installed_app_id=INSTALLED_APP_ID),
        response={'count': 1})

    # InstallApp Subscriptions
    mocker.get(API_SUBSCRIPTIONS.format(installed_app_id=INSTALLED_APP_ID),
               response='subscriptions_get_response')
    mocker.delete(API_SUBSCRIPTIONS.format(installed_app_id=INSTALLED_APP_ID),
                  response={'count': 3})
    mocker.post(API_SUBSCRIPTIONS.format(installed_app_id=INSTALLED_APP_ID),
                request='subscription_post_request',
                response='subscription_post_response')
    mocker.get(
        API_SUBSCRIPTION.format(
            installed_app_id=INSTALLED_APP_ID,
            subscription_id=SUBSCRIPTION_ID),
        response='subscription_capability_get_response')
    mocker.get(
        API_SUBSCRIPTION.format(
            installed_app_id=INSTALLED_APP_ID,
            subscription_id='498752fd-db87-4a5e-95f5-25a0e412838d'),
        response='subscription_device_get_response')
    mocker.delete(
        API_SUBSCRIPTION.format(
            installed_app_id=INSTALLED_APP_ID,
            subscription_id=SUBSCRIPTION_ID),
        response={'count': 1})

    # OAuth Token
    mocker.request('post', API_OAUTH_TOKEN, response='token_response')

    # Scenes
    mocker.get(API_SCENES, response='scenes')
    mocker.get(API_SCENES, response='scenes_location_filter',
               params=[('locationId', LOCATION_ID)])
    mocker.post(API_SCENE_EXECUTE.format(scene_id=SCENE_ID),
                response={'status': 'success'})


@pytest.fixture
def smartthings(event_loop):
    """Fixture for testing against the SmartThings class."""
    # Python 3.5 doesn't support yield in an async method so we have to
    # run the creation and clean-up of the session in the loop manually.
    mocker = ClientMocker()
    register_url_mocks(mocker)
    session = event_loop.run_until_complete(
        __create_session(event_loop, mocker))
    yield SmartThings(session, AUTH_TOKEN)
    event_loop.run_until_complete(session.close())


@pytest.fixture
def api(event_loop):
    """Fixture for testing against the API."""
    # Python 3.5 doesn't support yield in an async method so we have to
    # run the creation and clean-up of the session in the loop manually.
    mocker = ClientMocker()
    register_url_mocks(mocker)
    session = event_loop.run_until_complete(
        __create_session(event_loop, mocker))
    yield Api(session, AUTH_TOKEN)
    event_loop.run_until_complete(session.close())


async def __create_session(event_loop, mocker):
    return mocker.create_session(event_loop)



"""Tests for the installedapp module."""

import pytest

from pysmartthings.installedapp import (
    InstalledApp, InstalledAppEntity, InstalledAppStatus, InstalledAppType)

from .conftest import APP_ID, INSTALLED_APP_ID
from .utilities import get_json


class TestInstalledApp:
    """Tests for the InstalledApp class."""

    @staticmethod
    def test_init():
        """Tests the initialization."""
        # Arrange/Act
        app = InstalledApp()
        # Assert
        assert app.installed_app_type is InstalledAppType.UNKNOWN
        assert app.installed_app_status is InstalledAppStatus.UNKNOWN
        assert app.classifications == []

    @staticmethod
    def test_apply_data():
        """Tests the apply_data function."""
        # Arrange
        app = InstalledApp()
        data = get_json('installedapp_get_response.json')
        # Act
        app.apply_data(data)
        # Assert
        assert app.installed_app_id == INSTALLED_APP_ID
        assert app.installed_app_type == InstalledAppType.WEBHOOK_SMART_APP
        assert app.installed_app_status == InstalledAppStatus.PENDING
        assert app.display_name == 'pysmartthings'
        assert app.app_id == APP_ID
        assert app.reference_id is None
        assert app.location_id == '397678e5-9995-4a39-9d9f-ae6ba310236b'
        assert app.created_date == '2018-12-19T02:49:58Z'
        assert app.last_updated_date == '2018-12-19T02:49:58Z'
        assert app.classifications == ['AUTOMATION']


class TestInstalledAppEntity:
    """Tests for the InstalledAppEntity class."""

    @staticmethod
    @pytest.mark.asyncio
    async def test_refresh(api):
        """Tests the refresh method."""
        # Arrange
        app = InstalledAppEntity(
            api, installed_app_id=INSTALLED_APP_ID)
        # Act
        await app.refresh()
        # Assert
        assert app.app_id == APP_ID

    @staticmethod
    @pytest.mark.asyncio
    async def test_save(api):
        """Tests the refresh method."""
        # Arrange
        app = InstalledAppEntity(api)
        # Act/Assert
        with pytest.raises(NotImplementedError):
            await app.save()

    @staticmethod
    @pytest.mark.asyncio
    async def test_subscriptions(api):
        """Tests the subscriptions method."""
        # Arrange
        app = InstalledAppEntity(
            api, installed_app_id=INSTALLED_APP_ID)
        await app.refresh()
        # Act
        subscriptions = await app.subscriptions()
        # Assert
        assert len(subscriptions) == 3



"""Defines the entity module."""

from .api import Api


class Entity:
    """Define an entity from the SmartThings API."""

    def __init__(self, api: Api):
        """Initialize a new instance of the entity."""
        self._api = api

    async def refresh(self):
        """Retrieve the latest values from the API."""
        raise NotImplementedError

    async def save(self):
        """Update or create the entity."""
        raise NotImplementedError



"""Tests for the error module."""

from pysmartthings.errors import APIResponseError


class TestAPIResponseError:
    """Tests for the APIResponseError class."""

    @staticmethod
    def test_is_target_error():
        """Tests the initialization."""
        # Arrange/Act
        data = {
            "requestId": "8B66A345-03B0-477F-A8A6-1A1CF0277040",
            "error": {
                "code": "ConstraintViolationError",
                "message": "The request is malformed.",
                "details": [
                    {
                        "code": "TargetTimeoutError",
                        "target": "https://blah.blah/blah",
                        "message": "Upstream target timed out",
                        "details": []
                    }
                ]
            }
        }
        error = APIResponseError(None, None, data=data)
        # Assert
        assert error.is_target_error()
        assert error.raw_error_response == data
        assert error.request_id == '8B66A345-03B0-477F-A8A6-1A1CF0277040'
        assert error.error.code == 'ConstraintViolationError'
        assert error.error.message == 'The request is malformed.'
        assert len(error.error.details) == 1
        detail = error.error.details[0]
        assert detail.code == 'TargetTimeoutError'
        assert detail.target == "https://blah.blah/blah"
        assert detail.message == "Upstream target timed out"
        assert not detail.details

    @staticmethod
    def test_str():
        """Tests the initialization."""
        # Arrange/Act
        data = {
            "requestId": "8B66A345-03B0-477F-A8A6-1A1CF0277040"
        }
        error = APIResponseError(None, None, status=422,
                                 message="Unprocessable Entity", data=data)
        # Assert
        assert str(error) == 'Unprocessable Entity (422): ' \
            '{"requestId": "8B66A345-03B0-477F-A8A6-1A1CF0277040"}'



"""Utility for invoking the SmartThings Cloud API."""

from typing import Optional, Sequence

from aiohttp import BasicAuth, ClientSession

from .errors import APIInvalidGrant, APIResponseError

API_OAUTH_TOKEN = "https://auth-global.api.smartthings.com/oauth/token"
API_BASE = "https://api.smartthings.com/v1/"
API_LOCATIONS = "locations"
API_LOCATION = API_LOCATIONS + "/{location_id}"
API_ROOMS = "locations/{location_id}/rooms"
API_ROOM = "locations/{location_id}/rooms/{room_id}"
API_DEVICES = "devices"
API_DEVICE = API_DEVICES + "/{device_id}"
API_DEVICE_STATUS = "devices/{device_id}/status"
API_DEVICE_COMMAND = "devices/{device_id}/commands"
API_APPS = "apps"
API_APP = "apps/{app_id}"
API_APP_OAUTH = "apps/{app_id}/oauth"
API_APP_OAUTH_GENERATE = "apps/{app_id}/oauth/generate"
API_APP_SETTINGS = "apps/{app_id}/settings"
API_INSTALLEDAPPS = "installedapps"
API_INSTALLEDAPP = "installedapps/{installed_app_id}"
API_SUBSCRIPTIONS = API_INSTALLEDAPP + "/subscriptions"
API_SUBSCRIPTION = API_SUBSCRIPTIONS + "/{subscription_id}"
API_SCENES = "scenes"
API_SCENE_EXECUTE = "scenes/{scene_id}/execute"


class Api:
    """
    Wrapper around the SmartThings Cloud API operations.

    https://smartthings.developer.samsung.com/docs/api-ref/st-api.html
    """

    __slots__ = ['_session', '_token', '_api_base']

    def __init__(self, session: ClientSession, token: str, *,
                 api_base: str = API_BASE):
        """Create a new API with the given session and token."""
        self._session = session
        self._token = token
        self._api_base = api_base

    async def get_locations(self) -> dict:
        """
        Get locations.

        https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/listLocations
        """
        return await self.get_items(API_LOCATIONS)

    async def get_location(self, location_id: str) -> dict:
        """
        Get a specific location.

        https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/getLocation
        """
        return await self.get(API_LOCATION.format(location_id=location_id))

    async def get_rooms(self, location_id: str) -> dict:
        """
        Get a location's rooms.

        This API call is undocumented.
        """
        return await self.get_items(API_ROOMS.format(location_id=location_id))

    async def get_room(self, location_id: str, room_id: str) -> dict:
        """
        Get a specific room within a location.

        This API call is undocumented.
        """
        return await self.get(
            API_ROOM.format(location_id=location_id, room_id=room_id))

    async def create_room(self, location_id: str, data: dict):
        """
        Create a room.

        This API call is undocumented.
        """
        return await self.post(
            API_ROOMS.format(location_id=location_id), data)

    async def update_room(self, location_id: str, room_id: str, data: dict):
        """
        Update a room.

        This API call is undocumented.
        """
        return await self.put(
            API_ROOM.format(location_id=location_id, room_id=room_id), data)

    async def delete_room(self, location_id: str, room_id: str):
        """
        Delete a room.

        This API call is undocumented.
        """
        return await self.delete(
            API_ROOM.format(location_id=location_id, room_id=room_id))

    async def get_devices(self, params: Optional = None) -> dict:
        """
        Get the device definitions.

        https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/getDevices
        """
        return await self.get_items(API_DEVICES, params=params)

    async def get_device(self, device_id: str) -> dict:
        """
        Get as specific device.

        https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/getDevice
        """
        return await self.get(API_DEVICE.format(device_id=device_id))

    async def get_device_status(self, device_id: str) -> dict:
        """Get the status of a specific device."""
        return await self.get(API_DEVICE_STATUS.format(device_id=device_id))

    async def post_device_command(self, device_id, component_id, capability,
                                  command, args) -> object:
        """
        Execute commands on a device.

        https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/executeDeviceCommands
        """
        data = {
            "commands": [
                {
                    "component": component_id,
                    "capability": capability,
                    "command": command
                }
            ]
        }
        if args:
            data["commands"][0]["arguments"] = args

        return await self.post(
            API_DEVICE_COMMAND.format(device_id=device_id), data)

    async def get_apps(self, params: Optional = None) -> dict:
        """
        Get list of apps.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/listApps
        """
        return await self.get_items(API_APPS, params=params)

    async def get_app(self, app_id: str) -> dict:
        """
        Get the details of the specific app.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/getApp
        """
        return await self.get(API_APP.format(app_id=app_id))

    async def create_app(self, data: dict) -> dict:
        """
        Create a new app.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/createApp
        """
        return await self.post(API_APPS, data)

    async def update_app(self, app_id: str, data: dict) -> dict:
        """
        Update an existing app.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/updateApp
        """
        return await self.put(API_APP.format(app_id=app_id), data)

    async def delete_app(self, app_id: str):
        """
        Delete an app.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/deleteApp
        """
        return await self.delete(API_APP.format(app_id=app_id))

    async def get_app_settings(self, app_id: str) -> dict:
        """
        Get an app's settings.

        https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/getAppSettings
        """
        return await self.get(API_APP_SETTINGS.format(app_id=app_id))

    async def update_app_settings(self, app_id: str, data: dict) -> dict:
        """
        Update an app's settings.

        https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/updateAppSettings
        """
        return await self.put(API_APP_SETTINGS.format(app_id=app_id), data)

    async def get_app_oauth(self, app_id: str) -> dict:
        """
        Get an app's oauth settings.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/getAppOauth
        """
        return await self.get(API_APP_OAUTH.format(app_id=app_id))

    async def update_app_oauth(self, app_id: str, data: dict) -> dict:
        """
        Update an app's oauth settings.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/updateAppOauth
        """
        return await self.put(API_APP_OAUTH.format(app_id=app_id), data)

    async def generate_app_oauth(self, app_id: str, data: dict) -> dict:
        """
        Generate a new app oauth client/secret.

         https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/generateAppOauth
        """
        return await self.post(
            API_APP_OAUTH_GENERATE.format(app_id=app_id), data)

    async def get_installed_apps(self, params: Optional = None) -> dict:
        """
        Get list of installedapps.

        https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/listInstallations
        """
        return await self.get_items(API_INSTALLEDAPPS, params=params)

    async def get_installed_app(self, installed_app_id: str) -> dict:
        """
        Get the details of the specific installedapp.

        https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/getInstallation
        """
        return await self.get(
            API_INSTALLEDAPP.format(installed_app_id=installed_app_id))

    async def delete_installed_app(self, installed_app_id: str):
        """
        Delete an app.

        https://smartthings.developer.samsung.com/docs/api-ref/st-api.html#operation/deleteInstallation
        """
        return await self.delete(
            API_INSTALLEDAPP.format(installed_app_id=installed_app_id))

    async def get_subscriptions(self, installed_app_id: str) -> dict:
        """
        Get installedapp's subscriptions.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/listSubscriptions
        """
        return await self.get_items(
            API_SUBSCRIPTIONS.format(installed_app_id=installed_app_id))

    async def create_subscription(self, installed_app_id: str,
                                  data: dict) -> dict:
        """
        Create a subscription for an installedapp.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/saveSubscription
        """
        return await self.post(
            API_SUBSCRIPTIONS.format(installed_app_id=installed_app_id),
            data)

    async def delete_all_subscriptions(self, installed_app_id: str) -> dict:
        """
        Delete all subscriptions for an installedapp.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/deleteAllSubscriptions
        """
        return await self.delete(
            API_SUBSCRIPTIONS.format(installed_app_id=installed_app_id))

    async def get_subscription(self, installed_app_id: str,
                               subscription_id: str) -> dict:
        """
        Get an individual subscription.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/getSubscription
        """
        return await self.get(API_SUBSCRIPTION.format(
            installed_app_id=installed_app_id,
            subscription_id=subscription_id))

    async def delete_subscription(self, installed_app_id: str,
                                  subscription_id: str):
        """
        Delete an individual subscription.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/deleteSubscription
        """
        return await self.delete(API_SUBSCRIPTION.format(
            installed_app_id=installed_app_id,
            subscription_id=subscription_id))

    async def get_scenes(self, params: Optional = None):
        """
        Get scenes.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/listScenes
        """
        return await self.get_items(API_SCENES, params=params)

    async def execute_scene(self, scene_id: str) -> bool:
        """
        Execute a scene.

        https://smartthings.developer.samsung.com/develop/api-ref/st-api.html#operation/executeScene
        """
        return await self.post(
            API_SCENE_EXECUTE.format(scene_id=scene_id), data=None)

    @property
    def session(self) -> ClientSession:
        """Get the instance of the session."""
        return self._session

    @session.setter
    def session(self, value: ClientSession):
        """Set the instance of the session."""
        self._session = value

    @property
    def token(self):
        """Get the token used when making requests."""
        return self._token

    @token.setter
    def token(self, value):
        """Set the token to use when making requests."""
        self._token = value

    async def request(self, method: str, url: str, params: dict = None,
                      data: dict = None):
        """Perform a request against the specified parameters."""
        async with self._session.request(
                method, url, params=params, json=data,
                headers={"Authorization": "Bearer " + self._token}) as resp:
            if resp.status == 200:
                return await resp.json()
            if resp.status in (400, 422, 429, 500):
                data = None
                try:
                    data = await resp.json()
                except Exception:  # pylint: disable=broad-except
                    pass
                raise APIResponseError(
                    resp.request_info,
                    resp.history,
                    status=resp.status,
                    message=resp.reason,
                    headers=resp.headers,
                    data=data)
            resp.raise_for_status()

    async def get(self, resource: str, *, params: dict = None):
        """Get a resource."""
        return await self.request('get', self._api_base + resource, params)

    async def get_items(self, resource: str, *, params: dict = None):
        """Perform requests for a list of items that may have pages."""
        resp = await self.request(
            'get', self._api_base + resource, params, None)
        items = resp.get('items', [])
        next_link = Api._get_next_link(resp)
        while next_link:
            resp = await self.request('get', next_link, params, None)
            items.extend(resp.get('items', []))
            next_link = Api._get_next_link(resp)
        return items

    async def post(self, resource: str, data: Optional[Sequence]):
        """Perform a post request."""
        return await self.request('post', self._api_base + resource,
                                  data=data)

    async def put(self, resource: str, data: Optional[Sequence]):
        """Perform a put request."""
        return await self.request('put', self._api_base + resource,
                                  data=data)

    async def delete(self, resource: str, *, params: dict = None):
        """Delete a resource."""
        return await self.request('delete', self._api_base + resource, params)

    async def generate_tokens(self, client_id: str, client_secret: str,
                              refresh_token: str):
        """Obtain a new access and refresh token."""
        payload = {
            'grant_type': 'refresh_token',
            'refresh_token': refresh_token
        }
        async with self._session.request(
                'post', API_OAUTH_TOKEN,
                auth=BasicAuth(client_id, client_secret),
                data=payload) as resp:
            if resp.status == 200:
                return await resp.json()
            if resp.status == 400:
                data = {}
                try:
                    data = await resp.json()
                except Exception:  # pylint: disable=broad-except
                    pass
                raise APIInvalidGrant(data.get('error_description'))
            resp.raise_for_status()

    @staticmethod
    def _get_next_link(data):
        links = data.get('_links')
        if not links:
            return None
        next_link = links.get('next')
        if not next_link:
            return None
        return next_link.get('href')



"""Defines the rooms module."""
from typing import Dict, Optional

from .api import Api
from .entity import Entity


class Room:
    """Defines a SmartThings room."""

    def __init__(self):
        """Initialize the room."""
        self._room_id = None
        self._location_id = None
        self._name = None
        self._background_image = None

    def apply_data(self, data: dict):
        """Apply the data structure to the class."""
        self._room_id = data['roomId']
        self._location_id = data['locationId']
        self._name = data['name']
        self._background_image = data['backgroundImage']

    def to_data(self) -> dict:
        """Get a data structure representing this entity."""
        data = {
            'name': self._name,
            'backgroundImage': self._background_image,
        }
        return data

    @property
    def room_id(self) -> str:
        """Get the id of the room."""
        return self._room_id

    @room_id.setter
    def room_id(self, value: str):
        """Set the id of the room."""
        self._room_id = value

    @property
    def location_id(self) -> str:
        """Get the location id the room is part of."""
        return self._location_id

    @location_id.setter
    def location_id(self, value: str):
        """Set the location id of the room."""
        self._location_id = value

    @property
    def name(self) -> str:
        """Get the name of the room."""
        return self._name

    @name.setter
    def name(self, value: str):
        """Set the name of the room."""
        self._name = value

    @property
    def background_image(self) -> str:
        """Get the background image of the room."""
        return self._background_image

    @background_image.setter
    def background_image(self, value: str):
        """Set the background image."""
        self._background_image = value


class RoomEntity(Room, Entity):
    """Defines a SmartThings room entity."""

    def __init__(self, api: Api, data: Optional[Dict] = None, *,
                 location_id: str = None, room_id: str = None):
        """Initialize the room."""
        Entity.__init__(self, api)
        Room.__init__(self)
        if data:
            self.apply_data(data)
        if location_id:
            self._location_id = location_id
        if room_id:
            self._room_id = room_id

    async def refresh(self):
        """Refresh the room."""
        data = await self._api.get_room(self._location_id, self._room_id)
        if data:
            self.apply_data(data)

    async def save(self):
        """Save changes to the room."""
        data = await self._api.update_room(
            self._location_id, self._room_id, self.to_data())
        if data:
            self.apply_data(data)



"""Define the subscription module."""

from enum import Enum
from typing import Any, Optional

from .api import Api
from .entity import Entity


class SourceType(Enum):
    """Define the source type of a subscription."""

    UNKNOWN = 'UNKNOWN'
    DEVICE = 'DEVICE'
    CAPABILITY = 'CAPABILITY'


class Subscription:
    """Define the subscription class."""

    def __init__(self):
        """Initialize a new instance of the subscription class."""
        self._subscription_id = None
        self._installed_app_id = None
        self._source_type = SourceType.UNKNOWN
        self._capability = '*'
        self._attribute = '*'
        self._value = '*'
        self._state_change_only = True
        self._subscription_name = None
        # Capability-specific attributes
        self._location_id = None
        # Device-specific attributes
        self._device_id = None
        self._component_id = None

    def apply_data(self, data: dict):
        """Set the states of the app with the supplied data."""
        self._subscription_id = data['id']
        self._installed_app_id = data['installedAppId']
        self._source_type = SourceType(data['sourceType'])
        if self._source_type is SourceType.CAPABILITY:
            capability = data['capability']
            self._location_id = capability['locationId']
            self._capability = capability['capability']
            self._attribute = capability.get('attribute', '*')
            self._value = capability.get('value', '*')
            self._state_change_only = capability.get(
                'stateChangeOnly', True)
            self._subscription_name = capability.get(
                'subscriptionName', None)
        if self._source_type is SourceType.DEVICE:
            device = data['device']
            self._device_id = device['deviceId']
            self._component_id = device.get('componentId', '*')
            self._capability = device.get('capability', '*')
            self._attribute = device.get('attribute', '*')
            self._value = device.get('value', '*')
            self._state_change_only = device.get('stateChangeOnly', True)
            self._subscription_name = device.get('subscriptionName', None)

    def to_data(self) -> dict:
        """Get a data structure representing this entity."""
        data = {
            'sourceType': self._source_type.value
        }
        if self._source_type is SourceType.CAPABILITY:
            capability = {
                'locationId': self._location_id,
                'capability': self._capability
            }
            if self._attribute and self._attribute != '*':
                capability['attribute'] = self._attribute
            if self._value and self._value != '*':
                capability['value'] = self._value
            if not self._state_change_only:
                capability['stateChangeOnly'] = False
            if self._subscription_name:
                capability['subscriptionName'] = self._subscription_name
            data['capability'] = capability
        if self._source_type is SourceType.DEVICE:
            device = {
                'deviceId': self._device_id,
                'stateChangeOnly': self._state_change_only
            }
            if self._component_id and self._component_id != '*':
                device['componentId'] = self._component_id
            if self._capability and self._capability != '*':
                device['capability'] = self._capability
            if self._attribute and self._attribute != '*':
                device['attribute'] = self._attribute
            if self._value and self._value != '*':
                device['value'] = self._value
            if self._subscription_name:
                device['subscriptionName'] = self._subscription_name
            data['device'] = device
        return data

    @property
    def subscription_id(self) -> str:
        """Get the id of the subscription."""
        return self._subscription_id

    @property
    def installed_app_id(self) -> str:
        """Get the id of the subscribing app."""
        return self._installed_app_id

    @installed_app_id.setter
    def installed_app_id(self, value: str):
        """Set the id of the subscripting app."""
        self._installed_app_id = value

    @property
    def source_type(self) -> SourceType:
        """Get the type of the event that is being subscribed to."""
        return self._source_type

    @source_type.setter
    def source_type(self, value: Any):
        """Set the typ eof event that is being subscribed to."""
        self._source_type = SourceType(value)

    @property
    def capability(self) -> str:
        """Get the name of the capability that is subscribed."""
        return self._capability

    @capability.setter
    def capability(self, value: str):
        """Get the name of the capability that is subscribed."""
        self._capability = value

    @property
    def attribute(self) -> str:
        """Get the name of the capabilities attribute or * for all."""
        return self._attribute

    @attribute.setter
    def attribute(self, value: str):
        """Set the name of the capabilities attribute or * for all."""
        self._attribute = value

    @property
    def value(self) -> str:
        """Get the value for that will trigger the subscription."""
        return self._value

    @value.setter
    def value(self, value: str):
        """Set the value for that will trigger the subscription."""
        self._value = value

    @property
    def state_change_only(self) -> bool:
        """Get to execute only on a state change."""
        return self._state_change_only

    @state_change_only.setter
    def state_change_only(self, value: bool):
        """Set to execute only on a state change."""
        self._state_change_only = value

    @property
    def subscription_name(self) -> str:
        """Get a name for the subscription."""
        return self._subscription_name

    @subscription_name.setter
    def subscription_name(self, value: str):
        """Set a name for the subscription."""
        self._subscription_name = value

    @property
    def location_id(self) -> str:
        """Get the location id that both the app and source device are in."""
        return self._location_id

    @location_id.setter
    def location_id(self, value: str):
        """Set the location id that both the app and source device are in."""
        self._location_id = value

    @property
    def device_id(self):
        """Get the GUID of the device that is subscribed to."""
        return self._device_id

    @device_id.setter
    def device_id(self, value: str):
        """Set the GUID of the device that is subscribed to."""
        self._device_id = value

    @property
    def component_id(self) -> str:
        """Get the component ID on the device that is subscribed to."""
        return self._component_id

    @component_id.setter
    def component_id(self, value: str):
        """Set the component ID on the device that is subscribed to."""
        self._component_id = value


class SubscriptionEntity(Entity, Subscription):
    """Define a subscription entity."""

    def __init__(self, api: Api, data: Optional[dict] = None):
        """Create a new instance of the SubscriptionEntity class."""
        Entity.__init__(self, api)
        Subscription.__init__(self)
        if data:
            self.apply_data(data)

    async def refresh(self):
        """Refresh the subscription information using the API."""
        data = await self._api.get_subscription(
            self._installed_app_id, self._subscription_id)
        self.apply_data(data)

    async def save(self):
        """Subscriptions cannot be updated."""
        raise NotImplementedError



"""Define the oauth module."""

from datetime import datetime, timedelta
from typing import List, Optional

from .api import Api


class OAuthToken:
    """Define oauth token information."""

    def __init__(self, api: Api, data: Optional[dict] = None,
                 refresh_token: Optional[str] = None):
        """Create a new instance of the OAuthToken class."""
        self._api = api
        self._access_token = None
        self._refresh_token = refresh_token
        self._expires_in = 0
        self._token_type = None
        self._scope = []
        self._expiration_date = datetime.now()
        if data:
            self.apply_data(data)

    def apply_data(self, data: dict):
        """Apply the given data to the entity."""
        self._access_token = data['access_token']
        self._token_type = data['token_type']
        self._refresh_token = data['refresh_token']
        self._expires_in = data['expires_in']
        self._expiration_date = \
            datetime.now() + timedelta(0, self._expires_in)
        data_scope = data['scope']
        if isinstance(data_scope, list):
            self._scope = data_scope
        if isinstance(data_scope, str):
            self._scope.clear()
            self._scope.append(data_scope)

    async def refresh(self, client_id: str, client_secret: str):
        """Refresh the auth and refresh tokens."""
        data = await self._api.generate_tokens(client_id, client_secret,
                                               self._refresh_token)
        if data:
            self.apply_data(data)

    @property
    def access_token(self) -> str:
        """Get the access token for authentication."""
        return self._access_token

    @property
    def refresh_token(self) -> str:
        """Get the refresh token for obtaining new access tokens."""
        return self._refresh_token

    @property
    def expires_in(self) -> int:
        """Get the amount of time in seconds until the token expires."""
        return self._expires_in

    @property
    def token_type(self) -> str:
        """Get the type of token."""
        return self._token_type

    @property
    def scope(self) -> List[str]:
        """Get the scopes the token has permission to."""
        return self._scope

    @property
    def expiration_date(self) -> datetime:
        """Get the date when the token expires."""
        return self._expiration_date

    @property
    def is_expired(self):
        """Return True if the token has expired."""
        return datetime.now() > self._expiration_date



"""Tests for the subscription module."""

import pytest

from pysmartthings.subscription import (
    SourceType, Subscription, SubscriptionEntity)

from .conftest import INSTALLED_APP_ID
from .utilities import get_json


class TestSubscription:
    """Tests for the Subscription class."""

    @staticmethod
    def test_init():
        """Test the initialization method."""
        # Arrange/Act
        sub = Subscription()
        # Assert
        assert sub.source_type is SourceType.UNKNOWN
        assert sub.capability == '*'
        assert sub.attribute == '*'
        assert sub.value == '*'
        assert sub.state_change_only

    @staticmethod
    def test_apply_data_capability():
        """Test apply data."""
        # Arrange
        data = get_json('subscription_capability_get_response.json')
        sub = Subscription()
        # Act
        sub.apply_data(data)
        # Assert
        assert sub.subscription_id == '7bdf5909-57c4-41f3-9089-e520513bd92a'
        assert sub.installed_app_id == INSTALLED_APP_ID
        assert sub.source_type == SourceType.CAPABILITY
        assert sub.location_id == '397678e5-9995-4a39-9d9f-ae6ba310236b'
        assert sub.capability == 'switchLevel'
        assert sub.attribute == '*'
        assert sub.value == '*'
        assert sub.state_change_only
        assert sub.subscription_name == 'switchLevel_sub'

    @staticmethod
    def test_apply_data_device():
        """Test apply data."""
        # Arrange
        data = get_json('subscription_device_get_response.json')
        sub = Subscription()
        # Act
        sub.apply_data(data)
        # Assert
        assert sub.subscription_id == '498752fd-db87-4a5e-95f5-25a0e412838d'
        assert sub.installed_app_id == INSTALLED_APP_ID
        assert sub.source_type == SourceType.DEVICE
        assert sub.device_id == '64e7f664-5b99-4573-b76d-03be3021dc78'
        assert sub.component_id == '*'
        assert sub.capability == '*'
        assert sub.attribute == '*'
        assert sub.value == '*'
        assert sub.state_change_only
        assert not sub.subscription_name

    @staticmethod
    def test_to_data_capability():
        """Test the to_data method for capabilities."""
        # Arrange
        sub = Subscription()
        sub.source_type = SourceType.CAPABILITY
        sub.location_id = '397678e5-9995-4a39-9d9f-ae6ba310236b'
        sub.capability = 'switch'
        sub.attribute = 'switchLevel'
        sub.value = '100'
        sub.state_change_only = False
        sub.subscription_name = "Test"
        # Act
        data = sub.to_data()
        # Assert
        assert data['sourceType'] == SourceType.CAPABILITY.value
        assert data['capability']['locationId'] == \
            '397678e5-9995-4a39-9d9f-ae6ba310236b'
        assert data['capability']['capability'] == 'switch'
        assert data['capability']['attribute'] == 'switchLevel'
        assert data['capability']['value'] == '100'
        assert data['capability']['subscriptionName'] == 'Test'
        assert not data['capability']['stateChangeOnly']

    @staticmethod
    def test_to_data_device():
        """Test the to_data method for devices."""
        # Arrange
        sub = Subscription()
        sub.source_type = SourceType.DEVICE
        sub.device_id = '397678e5-9995-4a39-9d9f-ae6ba310236b'
        sub.component_id = 'main'
        sub.capability = 'switch'
        sub.attribute = 'switchLevel'
        sub.value = '100'
        sub.state_change_only = True
        sub.subscription_name = "Test"
        # Act
        data = sub.to_data()
        # Assert
        assert data['sourceType'] == SourceType.DEVICE.value
        assert data['device']['deviceId'] == \
            '397678e5-9995-4a39-9d9f-ae6ba310236b'
        assert data['device']['componentId'] == 'main'
        assert data['device']['capability'] == "switch"
        assert data['device']['attribute'] == "switchLevel"
        assert data['device']['value'] == "100"
        assert data['device']['subscriptionName'] == 'Test'
        assert data['device']['stateChangeOnly']


class TestSubscriptionEntity:
    """Tests for the SubscriptionEntity class."""

    @staticmethod
    @pytest.mark.asyncio
    async def test_refresh(api):
        """Tests the refresh method."""
        # Arrange
        app = SubscriptionEntity(api)
        app.apply_data({
            'id': '7bdf5909-57c4-41f3-9089-e520513bd92a',
            'installedAppId': INSTALLED_APP_ID,
            'sourceType': 'UNKNOWN'
        })
        # Act
        await app.refresh()
        # Assert
        assert app.subscription_name == 'switchLevel_sub'

    @staticmethod
    @pytest.mark.asyncio
    async def test_save(api):
        """Tests the refresh method."""
        # Arrange
        app = SubscriptionEntity(api)
        # Act/Assert
        with pytest.raises(NotImplementedError):
            await app.save()



"""Define the app module."""

import re
from typing import List, Optional

from .api import Api
from .entity import Entity

APP_TYPE_LAMBDA = "LAMBDA_SMART_APP"
APP_TYPE_WEBHOOK = "WEBHOOK_SMART_APP"
CLASSIFICATION_AUTOMATION = "AUTOMATION"

_APP_NAME_PATTERN = re.compile('^[a-z0-9._-]{1,250}$', re.IGNORECASE)


class App:
    """Define the app class."""

    def __init__(self):
        """Initialize a new instance of the App class."""
        self._app_name = None
        self._display_name = None
        self._description = None
        self._single_instance = False
        self._app_type = None
        self._classifications = []
        self._lambda_functions = []
        self._webhook_target_url = None
        self._webhook_public_key = None
        self._app_id = None
        self._created_date = None
        self._last_updated_date = None

    def apply_data(self, data: dict):
        """Set the states of the app with the supplied data."""
        self._app_name = data['appName']
        self._app_id = data['appId']
        self._app_type = data['appType']
        self._classifications = data['classifications']
        self._display_name = data['displayName']
        self._description = data['description']
        self._created_date = \
            data.get('createdDate', self._created_date)
        self._last_updated_date = \
            data.get('lastUpdatedDate', self._last_updated_date)
        self._single_instance = \
            data.get('singleInstance', self._single_instance)
        if self.app_type == APP_TYPE_WEBHOOK and 'webhookSmartApp' in data:
            self._webhook_target_url = data['webhookSmartApp']['targetUrl']
            self._webhook_public_key = data['webhookSmartApp']['publicKey']
        if self.app_type == APP_TYPE_LAMBDA and 'lambdaSmartApp' in data:
            self._lambda_functions = data['lambdaSmartApp']['functions']

    def to_data(self) -> dict:
        """Get a data structure representing this entity."""
        data = {
            'appName': self._app_name,
            'displayName': self._display_name,
            'description': self._description,
            'singleInstance': self._single_instance,
            'classifications': self._classifications,
            'appType': self._app_type
        }
        if self._app_type == APP_TYPE_WEBHOOK:
            data['webhookSmartApp'] = {
                'targetUrl': self._webhook_target_url
            }
        if self._app_type == APP_TYPE_LAMBDA:
            data['lambdaSmartApp'] = {
                'functions': self._lambda_functions
            }
        return data

    @property
    def app_id(self) -> str:
        """Get the application id."""
        return self._app_id

    @property
    def created_date(self):
        """Get the created date of the app."""
        return self._created_date

    @property
    def last_updated_date(self):
        """Get the last updated date of the app."""
        return self._last_updated_date

    @property
    def app_name(self) -> str:
        """
        Get the app name.

        A globally unique, developer-defined identifier for an app. It is
        alpha-numeric, may contain dashes, underscores, periods, and must
        be less then 250 characters long. ^[a-z0-9.-_]{1,250}$
        """
        return self._app_name

    @app_name.setter
    def app_name(self, value: str):
        """Set the app name."""
        if not value:
            raise ValueError("value cannot be None or zero length.")
        if not _APP_NAME_PATTERN.match(value):
            raise ValueError("value must be alpha-numeric, may contain dashes "
                             "underscores, periods, and must be less then 250 "
                             "characters long.")
        self._app_name = value

    @property
    def display_name(self) -> str:
        """
        Get the display name.

        A default display name for an app. <= 75 characters
        """
        return self._display_name

    @display_name.setter
    def display_name(self, value: str):
        """Set the display name."""
        if not value:
            raise ValueError("value cannot be None or zero length.")
        if len(value) > 75:
            raise ValueError("value must be <= 75 characters in length.")
        self._display_name = value

    @property
    def description(self) -> str:
        """
        Get the description.

        A default description for an app. <= 250 characters
        """
        return self._description

    @description.setter
    def description(self, value: str):
        """Set the description."""
        if not value:
            raise ValueError("value cannot be None or zero length.")
        if len(value) > 250:
            raise ValueError("value must be <= 250 characters in length.")
        self._description = value

    @property
    def single_instance(self) -> bool:
        """
        Get the single instance parameter.

        Inform the installation systems that a particular app can only be
        installed once within a user's account.
        """
        return self._single_instance

    @single_instance.setter
    def single_instance(self, value: bool):
        """Set the single instance parameter."""
        self._single_instance = bool(value)

    @property
    def classifications(self) -> List[str]:
        """Get the classifications of the app."""
        return self._classifications

    @classifications.setter
    def classifications(self, value: List[str]):
        """Set the classifications of the app."""
        self._classifications = list(value)

    @property
    def app_type(self) -> str:
        """
        Get the app type.

        Denotes the type of app. "LAMBDA_SMART_APP" "WEBHOOK_SMART_APP"
        """
        return self._app_type

    @app_type.setter
    def app_type(self, value: str):
        """Set the app type."""
        if value not in (APP_TYPE_LAMBDA, APP_TYPE_WEBHOOK):
            raise ValueError("value must be 'LAMBDA_SMART_APP' "
                             "or 'WEBHOOK_SMART_APP'")
        self._app_type = value

    @property
    def lambda_functions(self) -> List[str]:
        """
        Get the list of AWS arns referencing a Lambda function.

        Details related to a Lambda Smart App implementation. This model
        should only be specified for apps of type LAMBDA_SMART_APP.
        """
        return self._lambda_functions

    @property
    def webhook_target_url(self) -> str:
        """Get the URL that should be invoked during execution."""
        return self._webhook_target_url

    @webhook_target_url.setter
    def webhook_target_url(self, value: str):
        """Set the URL that should be invoked during execution."""
        self._webhook_target_url = value

    @property
    def webhook_public_key(self) -> str:
        """Get the public half of an RSA key pair."""
        return self._webhook_public_key


class AppSettings:
    """Define a SmartThings app settings."""

    def __init__(self, app_id: str):
        """Create a new instance of the AppSettings class."""
        self._app_id = app_id
        self._settings = {}

    def apply_data(self, data: dict):
        """Set the states of the app with the supplied data."""
        self._settings = data.get('settings', {})

    def to_data(self) -> dict:
        """Get a data structure representing the entity."""
        return {
            'settings': self._settings
        }

    @property
    def app_id(self):
        """Get the associated app id."""
        return self._app_id

    @property
    def settings(self) -> dict:
        """Get the settings for the app."""
        return self._settings

    @settings.setter
    def settings(self, value: dict):
        """Set the settings for the app."""
        self._settings = value


class AppSettingsEntity(Entity, AppSettings):
    """Define a SmartThings App settings entity."""

    def __init__(self, api: Api, app_id: str, data=None):
        """Create a new instance of the AppSettingEntity class."""
        Entity.__init__(self, api)
        AppSettings.__init__(self, app_id)
        if data:
            self.apply_data(data)

    async def refresh(self):
        """Refresh the value of the entity."""
        if not self._app_id:
            raise ValueError("Cannot refresh without an app_id")
        data = await self._api.get_app_settings(self._app_id)
        self.apply_data(data)

    async def save(self):
        """Save the value of the entity."""
        if not self._app_id:
            raise ValueError("Cannot save without an app_id")
        data = await self._api.update_app_settings(
            self._app_id, self.to_data())
        self.apply_data(data)


class AppOAuth:
    """Define the app OAuth settings."""

    def __init__(self, app_id: str):
        """Initialize a new instance of the OAuth class."""
        self._app_id = app_id
        self._client_name = None
        self._scope = []

    def to_data(self) -> dict:
        """Get a data representation of the instance."""
        return {
            'clientName': self._client_name,
            'scope': self._scope
        }

    def apply_data(self, data: dict):
        """Load the data of the instance."""
        self._client_name = data['clientName']
        self._scope = data.get('scope', self._scope)

    @property
    def app_id(self) -> str:
        """Get the app id the settings are associated with."""
        return self._app_id

    @property
    def client_name(self) -> str:
        """Get the name given to the OAuth Client."""
        return self._client_name

    @client_name.setter
    def client_name(self, value: str):
        """Set the name given to the OAuth client."""
        if not value:
            raise ValueError("Value can not be None or an empty string.")
        self._client_name = value

    @property
    def scope(self) -> List[str]:
        """Get the list of SmartThings API OAuth scope identifiers."""
        return self._scope


class AppOAuthEntity(Entity, AppOAuth):
    """Define oauth client settings."""

    def __init__(self, api: Api, app_id: str, data=None):
        """Create a new instance of the OAuth class."""
        Entity.__init__(self, api)
        AppOAuth.__init__(self, app_id)
        if data:
            self.apply_data(data)

    async def refresh(self):
        """Retrieve the latest values from the API."""
        data = await self._api.get_app_oauth(self._app_id)
        if data:
            self.apply_data(data)

    async def save(self):
        """Save changes to the app OAuth Client settings."""
        response = await self._api.update_app_oauth(
            self._app_id, self.to_data())
        if response:
            self.apply_data(response)


class AppEntity(Entity, App):
    """Define a SmartThings App entity."""

    def __init__(self, api: Api, data=None):
        """Create a new instance of the AppEntity class."""
        Entity.__init__(self, api)
        App.__init__(self)
        if data:
            self.apply_data(data)

    async def refresh(self):
        """Refresh the app information using the API."""
        data = await self._api.get_app(self._app_id)
        self.apply_data(data)

    async def save(self):
        """Save the changes made to the app."""
        response = await self._api.update_app(self._app_id, self.to_data())
        self.apply_data(response)

    async def oauth(self) -> AppOAuthEntity:
        """Get the app's OAuth settings."""
        entity = await self._api.get_app_oauth(self._app_id)
        return AppOAuthEntity(self._api, self._app_id, entity)

    async def settings(self) -> AppSettingsEntity:
        """Get the app's settings."""
        entity = await self._api.get_app_settings(self._app_id)
        return AppSettingsEntity(self._api, self._app_id, entity)


class AppOAuthClient:
    """Define an oauth client information."""

    def __init__(self, data: Optional[dict]):
        """Create a new instance of the OAuthClient."""
        self._client_id = None
        self._client_secret = None
        if data:
            self.apply_data(data)

    def apply_data(self, data: dict):
        """Apply the given data to the entity."""
        self._client_id = data['oauthClientId']
        self._client_secret = data['oauthClientSecret']

    @property
    def client_id(self):
        """Get the client id."""
        return self._client_id

    @property
    def client_secret(self):
        """Get the client secret."""
        return self._client_secret


class AppOAuthClientEntity(AppOAuthClient):
    """Define an oauth client information details."""

    def __init__(self, api: Api, app_id: str, data: Optional[dict]):
        """Init the class."""
        self._client_details = AppOAuthEntity(api, app_id, None)
        super().__init__(data)

    def apply_data(self, data: dict):
        """Apply the given data to the entity."""
        super().apply_data(data)
        self._client_details.apply_data(data['oauthClientDetails'])

    @property
    def client_details(self) -> AppOAuthEntity:
        """Get the OAuth entity."""
        return self._client_details



"""Define testing utilities."""
import json as _json
from typing import Optional, Sequence, Union
from urllib.parse import parse_qs

from aiohttp import ClientSession
from aiohttp.client_exceptions import ClientResponseError
from yarl import URL

BodyFixtureType = Optional[Union[str, list, dict]]
BodyType = Optional[Union[list, dict]]


def get_json(file):
    """Load a json file."""
    with open("tests/json/" + file, "r") as json_file:
        return _json.load(json_file)


def _get_json_fixture(body: BodyFixtureType) -> BodyType:
    if isinstance(body, (dict, list)):
        return body
    if isinstance(body, str):
        return get_json(body + '.json')
    return None


class ClientMocker:
    """Mock Aiohttp client requests."""

    def __init__(self):
        """Initialize the request mocker."""
        self._mocks = []
        self.default_headers = None
        self.base_url = None

    def get(self, resource: str, *, params=None, response=None):
        """Register a mock get request."""
        self.request('get', self.base_url + resource, params=params,
                     headers=self.default_headers, response=response)

    def post(self, resource: str, *, params=None, request=None, response=None):
        """Register a mock post request."""
        self.request('post', self.base_url + resource, params=params,
                     headers=self.default_headers, request=request,
                     response=response)

    def put(self, resource: str, *, params=None, request=None, response=None):
        """Register a mock post request."""
        self.request('put', self.base_url + resource, params=params,
                     headers=self.default_headers, request=request,
                     response=response)

    def delete(self, resource: str, *, params=None, response=None):
        """Register a mock get request."""
        self.request('delete', self.base_url + resource, params=params,
                     headers=self.default_headers, response=response)

    def request(self, method: str, url: str, *,
                params=None, status=200, headers=None,
                request=None, response=None):
        """Register a mock request."""
        self._mocks.append(MockResponse(
            method, url, params, status, headers, request, response))

    def create_session(self, loop):
        """Create a ClientSession that is bound to this mocker."""
        session = ClientSession(loop=loop)
        # Setting directly on `session` will raise deprecation warning
        object.__setattr__(session, '_request', self.match_request)
        return session

    async def match_request(self, method, url, *, data=None, auth=None,
                            params=None, headers=None, allow_redirects=None,
                            timeout=None, json=None):
        """Match a request against pre-registered requests."""
        url = URL(url)
        if params:
            url = url.with_query(params)

        for response in self._mocks:
            if response.match_request(method, url, headers or [], json):
                return response

        assert False, "No mock registered for {} {} {}".format(method.upper(),
                                                               url, params)


class MockResponse:
    """Mock Aiohttp client response."""

    def __init__(self, method, url, params, status, headers, request,
                 response):
        """Initialize a fake response."""
        self.method = method
        url = URL(url)
        if params:
            url = url.with_query(params)
        self._url = url
        self.status = status
        self._response = response
        self._request = request
        self._headers = headers or []

    def match_request(self, method: str, url: URL,
                      headers: Optional[Sequence], json: BodyFixtureType):
        """Test if response answers request."""
        # Headers
        if self._headers != headers:
            return False
        # Method
        if method.lower() != self.method.lower():
            return False
        # Url
        if (self._url.scheme != url.scheme or self._url.host != url.host or
                self._url.path != url.path):
            return False
        # Query string
        request_qs = parse_qs(url.query_string)
        matcher_qs = parse_qs(self._url.query_string)
        if request_qs != matcher_qs:
            return False

        # Request body
        if self._request and not json == _get_json_fixture(self._request):
            return False
        return True

    @property
    def headers(self):
        """Return content_type."""
        return self._headers

    @property
    def cookies(self):
        """Return dict of cookies."""
        return {}

    @property
    def url(self):
        """Return yarl of URL."""
        return self._url

    @property
    def content(self):
        """Return content."""
        raise NotImplementedError

    async def read(self):
        """Return mock response."""
        raise NotImplementedError

    async def text(self, encoding='utf-8'):
        """Return mock response as a string."""
        raise NotImplementedError

    async def json(self, encoding='utf-8'):
        """Return mock response as a json."""
        return _get_json_fixture(self._response)

    def release(self):
        """Mock release."""
        pass

    def raise_for_status(self):
        """Raise error if status is 400 or higher."""
        if self.status >= 400:
            raise ClientResponseError(
                None, None, code=self.status, headers=self.headers)

    def close(self):
        """Mock close."""
        pass



"""Tests for the SmartThings file."""

import pytest

from pysmartthings.app import App, AppOAuth, AppSettings
from pysmartthings.room import Room
from pysmartthings.subscription import Subscription

from .conftest import (
    APP_ID, CLIENT_ID, CLIENT_SECRET, DEVICE_ID, INSTALLED_APP_ID, LOCATION_ID,
    REFRESH_TOKEN, ROOM_ID, SCENE_ID, SUBSCRIPTION_ID)
from .utilities import get_json


class TestSmartThings:
    """Tests for the SmartThings class."""

    @staticmethod
    @pytest.mark.asyncio
    async def test_devices(smartthings):
        """Tests devices are retrieved."""
        # Act
        devices = await smartthings.devices()
        # Assert
        assert len(devices) == 5

    @staticmethod
    @pytest.mark.asyncio
    async def test_devices_with_filter(smartthings):
        """Tests retrieving a filtered view of devices."""
        # Act
        devices = await smartthings.devices(
            location_ids=[LOCATION_ID],
            capabilities=['switch'],
            device_ids=['edd26ac6-d156-4505-9647-3b20118ae4d1',
                        'be1a61ce-c2a4-4b32-bf8c-31de6d3fa7dd'])
        # Assert
        assert len(devices) == 2

    @staticmethod
    @pytest.mark.asyncio
    async def test_device(smartthings):
        """Tests the device(id) method."""
        # Act
        device = await smartthings.device(DEVICE_ID)
        # Assert
        assert device.device_id == DEVICE_ID

    @staticmethod
    @pytest.mark.asyncio
    async def test_locations(smartthings):
        """Tests locations are retrieved."""
        # Act
        locations = await smartthings.locations()
        # Assert
        assert len(locations) == 2

    @staticmethod
    @pytest.mark.asyncio
    async def test_location(smartthings):
        """Tests the location(id) method."""
        # Act
        location = await smartthings.location(LOCATION_ID)
        # Assert
        assert location.location_id == LOCATION_ID

    @staticmethod
    @pytest.mark.asyncio
    async def test_rooms(smartthings):
        """Tests the rooms(id) method."""
        # Act
        rooms = await smartthings.rooms(LOCATION_ID)
        # Assert
        assert len(rooms) == 1

    @staticmethod
    @pytest.mark.asyncio
    async def test_create_room(smartthings):
        """Tests the create room method."""
        # Arrange
        room = Room()
        room.name = 'Theater'
        room.background_image = 'Test'
        room.location_id = LOCATION_ID
        # Act
        room = await smartthings.create_room(room)
        # Assert
        assert room.room_id == ROOM_ID

    @staticmethod
    @pytest.mark.asyncio
    async def test_update_room(smartthings):
        """Tests the create room method."""
        # Arrange
        room = Room()
        room.name = 'Theater'
        room.background_image = 'Test'
        room.location_id = LOCATION_ID
        room.room_id = ROOM_ID
        # Act
        room = await smartthings.update_room(room)
        # Assert
        assert room.room_id == ROOM_ID

    @staticmethod
    @pytest.mark.asyncio
    async def test_room(smartthings):
        """Tests the room(id, id) method."""
        # Act
        room = await smartthings.room(LOCATION_ID, ROOM_ID)
        # Assert
        assert room.room_id == ROOM_ID

    @staticmethod
    @pytest.mark.asyncio
    async def test_delete_room(smartthings):
        """Tests the delete room."""
        # Act
        result = await smartthings.delete_room(LOCATION_ID, ROOM_ID)
        # Assert
        assert result

    @staticmethod
    @pytest.mark.asyncio
    async def test_apps(smartthings):
        """Tests locations are retrieved."""
        # Act
        apps = await smartthings.apps()
        # assert
        assert len(apps) == 1

    @staticmethod
    @pytest.mark.asyncio
    async def test_app(smartthings):
        """Tests the app(id) method."""
        # Act
        app = await smartthings.app(APP_ID)
        # Assert
        assert app.app_id == APP_ID

    @staticmethod
    @pytest.mark.asyncio
    async def test_create_app(smartthings):
        """Tests the create app method."""
        # Arrange
        app = App()
        data = get_json('app_post_request.json')
        data['appId'] = APP_ID
        app.apply_data(data)
        # Act
        app, oauth = await smartthings.create_app(app)
        # Assert
        assert app.app_id == APP_ID
        assert oauth.client_id == '7cd4d474-7b36-4e03-bbdb-4cd4ae45a2be'
        assert oauth.client_secret == '9b3fd445-42d6-441b-b386-99ea51e13cb0'

    @staticmethod
    @pytest.mark.asyncio
    async def test_delete_app(smartthings):
        """Tests the delete app method."""
        # Act/Assert
        result = await smartthings.delete_app(APP_ID)
        # Assert
        assert result

    @staticmethod
    @pytest.mark.asyncio
    async def test_app_settings(smartthings):
        """Tests retrieval of app settings."""
        # Act
        settings = await smartthings.app_settings(APP_ID)
        # Assert
        assert settings.app_id == APP_ID
        assert settings.settings == {'test': 'test'}

    @staticmethod
    @pytest.mark.asyncio
    async def test_update_app_settings(smartthings):
        """Tests updating app settings."""
        # Arrange
        settings = AppSettings(APP_ID)
        settings.settings['test'] = 'test'
        # Act
        entity = await smartthings.update_app_settings(settings)
        # Assert
        assert entity.app_id == settings.app_id
        assert entity.settings == {'test': 'test'}

    @staticmethod
    @pytest.mark.asyncio
    async def test_app_oauth(smartthings):
        """Tests retrieval of OAuth settings."""
        # Act
        oauth = await smartthings.app_oauth(APP_ID)
        # Assert
        assert oauth.app_id == APP_ID
        assert oauth.client_name == 'pysmartthings-test'
        assert oauth.scope == ['r:devices']

    @staticmethod
    @pytest.mark.asyncio
    async def test_update_app_oauth(smartthings):
        """Tests updating OAuth settings."""
        # Arrange
        oauth = AppOAuth(APP_ID)
        oauth.client_name = 'pysmartthings-test'
        oauth.scope.append('r:devices')
        # Act
        oauth_entity = await smartthings.update_app_oauth(oauth)
        # Assert
        assert oauth_entity.app_id == oauth.app_id
        assert oauth_entity.client_name == oauth.client_name
        assert oauth_entity.scope == oauth.scope

    @staticmethod
    @pytest.mark.asyncio
    async def test_generate_app_oauth(smartthings):
        """Tests generating new OAuth info."""
        # Arrange
        oauth = AppOAuth(APP_ID)
        oauth.client_name = 'pysmartthings'
        oauth.scope.append('r:devices:*')
        # Act
        entity = await smartthings.generate_app_oauth(oauth)
        # Assert
        assert entity.client_id == '0b6a77d4-2dff-4b00-ba33-35f660fbfb83'
        assert entity.client_secret == '05a49aac-55d6-4092-96bf-7ca6ca3666d3'
        assert entity.client_details.app_id == APP_ID
        assert entity.client_details.client_name == 'pysmartthings'
        assert entity.client_details.scope == ['r:devices:$', 'r:devices:*']

    @staticmethod
    @pytest.mark.asyncio
    async def test_installed_apps(smartthings):
        """Tests the installedapps method."""
        # Act
        apps = await smartthings.installed_apps()
        # Assert
        assert len(apps) == 2

    @staticmethod
    @pytest.mark.asyncio
    async def test_installed_app(smartthings):
        """Tests the installedapp(id) method."""
        # Act
        app = await smartthings.installed_app(INSTALLED_APP_ID)
        # Assert
        assert app.installed_app_id == INSTALLED_APP_ID

    @staticmethod
    @pytest.mark.asyncio
    async def test_delete_installed_app(smartthings):
        """Tests the delete app method."""
        # Act/Assert
        result = await smartthings.delete_installed_app(
            INSTALLED_APP_ID)
        # Assert
        assert result

    @staticmethod
    @pytest.mark.asyncio
    async def test_subscriptions(smartthings):
        """Tests the get subscriptions method."""
        # Act
        subscriptions = await smartthings.subscriptions(INSTALLED_APP_ID)
        # Assert
        assert len(subscriptions) == 3

    @staticmethod
    @pytest.mark.asyncio
    async def test_delete_subscriptions(smartthings):
        """Tests the delete subscriptions method."""
        # Act
        count = await smartthings.delete_subscriptions(INSTALLED_APP_ID)
        # Assert
        assert count == 3

    @staticmethod
    @pytest.mark.asyncio
    async def test_delete_subscription(smartthings):
        """Tests the delete subscription method."""
        # Act
        deleted = await smartthings.delete_subscription(
            INSTALLED_APP_ID, SUBSCRIPTION_ID)
        # Assert
        assert deleted

    @staticmethod
    @pytest.mark.asyncio
    async def test_create_subscription(smartthings):
        """Tests the create subscription method."""
        # Arrange
        sub = Subscription()
        sub.source_type = 'CAPABILITY'
        sub.location_id = LOCATION_ID
        sub.capability = 'switch'
        sub.installed_app_id = INSTALLED_APP_ID
        # Act
        entity = await smartthings.create_subscription(sub)
        # Assert
        assert entity.subscription_id == SUBSCRIPTION_ID

    @staticmethod
    @pytest.mark.asyncio
    async def test_scenes(smartthings):
        """Tests the scenes method."""
        # Act
        scenes = await smartthings.scenes()
        # Assert
        assert len(scenes) == 2

    @staticmethod
    @pytest.mark.asyncio
    async def test_scenes_with_location_filter(smartthings):
        """Tests the scenes method."""
        # Act
        scenes = await smartthings.scenes(location_id=LOCATION_ID)
        # Assert
        assert len(scenes) == 1

    @staticmethod
    @pytest.mark.asyncio
    async def test_execute_scene(smartthings):
        """Tests the execute scene method."""
        # Act
        result = await smartthings.execute_scene(SCENE_ID)
        # Assert
        assert result

    @staticmethod
    @pytest.mark.asyncio
    async def test_generate_tokens(smartthings):
        """Tests the generate_tokens method."""
        # Act
        token = await smartthings.generate_tokens(
            CLIENT_ID, CLIENT_SECRET, REFRESH_TOKEN)
        # Assert
        assert token.refresh_token == '3d1a8d0a-a312-45c2-a9f5-95e59dc0e879'
        assert token.access_token == 'ad0fbf27-48d4-4ee9-ba47-7f5fedd7be35'
        assert token.scope == ['r:devices:*']
        assert token.token_type == 'bearer'
        assert token.expires_in == 299
        assert not token.is_expired



"""Tests for the OAuth module."""

import pytest

from pysmartthings.oauthtoken import OAuthToken

from .conftest import CLIENT_ID, CLIENT_SECRET, REFRESH_TOKEN
from .utilities import get_json


class TestOAuthToken:
    """Tests for the OAuthToken class."""

    @staticmethod
    def test_init():
        """Tests the init method."""
        # Arrange/Act
        token = OAuthToken(None, None)
        # Assert
        assert token.expires_in == 0
        assert token.scope == []

    @staticmethod
    def test_apply_data():
        """Tests the apply data method."""
        # Arrange
        data = get_json('token_response.json')
        # Act
        token = OAuthToken(None, data)
        # Assert
        assert token.expires_in == 299
        assert token.refresh_token == '3d1a8d0a-a312-45c2-a9f5-95e59dc0e879'
        assert token.access_token == 'ad0fbf27-48d4-4ee9-ba47-7f5fedd7be35'
        assert token.token_type == 'bearer'
        assert token.scope == ['r:devices:*']

    @staticmethod
    @pytest.mark.asyncio
    async def test_refresh(api):
        """Tests the refresh method."""
        # Arrange
        token = OAuthToken(api, refresh_token=REFRESH_TOKEN)
        # Act
        await token.refresh(CLIENT_ID, CLIENT_SECRET)
        # Assert
        assert token.refresh_token == '3d1a8d0a-a312-45c2-a9f5-95e59dc0e879'



"""Define tests for the app module."""

import pytest

from pysmartthings.app import (
    APP_TYPE_LAMBDA, APP_TYPE_WEBHOOK, CLASSIFICATION_AUTOMATION, App,
    AppEntity, AppOAuth, AppOAuthEntity, AppSettings, AppSettingsEntity)

from .conftest import APP_ID
from .utilities import get_json


class TestApp:
    """Define tests for the App class."""

    @staticmethod
    def test_init():
        """Tests the init function."""
        # Arrange/Act
        app = App()
        # Assert
        assert app.classifications is not None
        assert app.lambda_functions is not None

    @staticmethod
    def test_apply_data():
        """Tests the apply_data function."""
        # Arrange
        app = App()
        data = get_json('app_get.json')
        # Act
        app.apply_data(data)
        # Assert
        assert app.app_id == "c6cde2b0-203e-44cf-a510-3b3ed4706996"
        assert app.app_name == "pysmartthings-test"
        assert app.app_type == "WEBHOOK_SMART_APP"
        assert app.classifications == [CLASSIFICATION_AUTOMATION]
        assert app.display_name == "Test"
        assert app.description == \
            "A SmartApp that relays events to the pysmartthings library"
        assert app.single_instance
        assert app.webhook_target_url == \
            "https://homeassistant.sayre.net:8321/"
        assert app.webhook_public_key
        assert app.created_date == "2018-12-15T17:07:41Z"
        assert app.last_updated_date == "2018-12-15T17:07:42Z"

    @staticmethod
    def test_app_name():
        """Tests get/set of app_name."""
        # Arrange
        app = App()
        # Act
        expected_app_name = "my_app"
        app.app_name = expected_app_name
        app_name = app.app_name
        # Assert
        assert app_name == expected_app_name

    @staticmethod
    def test_app_name_invalid():
        """Tests valid values of app_name."""
        # Arrange
        app = App()
        # Act/Assert
        with pytest.raises(ValueError):
            app.app_name = ''
        with pytest.raises(ValueError):
            app.app_name = None
        with pytest.raises(ValueError):
            app.app_name = "My Super Cool App"

    @staticmethod
    def test_display_name():
        """Tests get/set of display_name."""
        # Arrange
        app = App()
        # Act
        expected = "My Display Name"
        app.display_name = expected
        actual = app.display_name
        # Assert
        assert expected == actual

    @staticmethod
    def test_display_name_invalid():
        """Tests valid values of display_name."""
        # Arrange
        app = App()
        # Act/Assert
        with pytest.raises(ValueError):
            app.display_name = ''
        with pytest.raises(ValueError):
            app.display_name = None
        with pytest.raises(ValueError):
            app.display_name = 'x' * 76

    @staticmethod
    def test_description():
        """Tests get/set of description."""
        # Arrange
        app = App()
        # Act
        expected = "My Description"
        app.description = expected
        actual = app.description
        # Assert
        assert expected == actual

    @staticmethod
    def test_description_invalid():
        """Tests valid values of description."""
        # Arrange
        app = App()
        # Act/Assert
        with pytest.raises(ValueError):
            app.description = ''
        with pytest.raises(ValueError):
            app.description = None
        with pytest.raises(ValueError):
            app.description = 'x' * 251

    @staticmethod
    def test_single_instance():
        """Tests get/set of single_instance."""
        # Arrange
        app = App()
        # Act
        expected = True
        app.single_instance = expected
        actual = app.single_instance
        # Assert
        assert expected == actual

    @staticmethod
    def test_app_type_webhook():
        """Tests get/set of app_type to webhook."""
        # Arrange
        app = App()
        # Act
        expected = APP_TYPE_WEBHOOK
        app.app_type = expected
        actual = app.app_type
        # Assert
        assert expected == actual

    @staticmethod
    def test_app_type_lambda():
        """Tests get/set of app_type to labmda."""
        # Arrange
        app = App()
        # Act
        expected = APP_TYPE_LAMBDA
        app.app_type = expected
        actual = app.app_type
        # Assert
        assert expected == actual

    @staticmethod
    def test_app_type_invalid():
        """Tests valid values of app_type."""
        # Arrange
        app = App()
        # Act/Assert
        with pytest.raises(ValueError):
            app.app_type = ''
        with pytest.raises(ValueError):
            app.app_type = None
        with pytest.raises(ValueError):
            app.app_type = 'Some type'

    @staticmethod
    def test_lambda_functions():
        """Tests get of lambda_functions."""
        # Arrange
        app = App()
        # Act
        app.lambda_functions.append(
            "arn:aws:lambda:eu-central-1:account-id:"
            "function:function-name:alias-name")
        # Assert
        assert app.lambda_functions

    @staticmethod
    def test_webhook_target_url():
        """Tests get/set of webhook_target_url."""
        # Arrange
        app = App()
        # Act
        expected = "http://my.web.site/"
        app.webhook_target_url = expected
        actual = app.webhook_target_url
        # Assert
        assert expected == actual


class TestAppSettings:
    """Tests for the AppSettings class."""

    @staticmethod
    def test_init():
        """Tests the init method."""
        # Arrange/Act
        settings = AppSettings(APP_ID)
        # Assert
        assert settings.app_id == APP_ID

    @staticmethod
    def test_apply_data():
        """Tests the apply_data method."""
        # Arrange
        data = get_json('app_settings.json')
        settings = AppSettings(APP_ID)
        # Act
        settings.apply_data(data)
        # Assert
        assert settings.settings['test'] == 'test'

    @staticmethod
    def test_to_data():
        """Tests the to_data method."""
        # Arrange
        settings = AppSettings(APP_ID)
        settings.settings = {'test': 'test'}
        # Act
        data = settings.to_data()
        # Assert
        assert data == {'settings': {'test': 'test'}}


class TestAppSettingsEntity:
    """Tests for the AppSettingsEntity class."""

    @staticmethod
    @pytest.mark.asyncio
    async def test_refresh(api):
        """Tests data is refreshed."""
        # Arrange
        data = {'settings': {'test2': 'test'}}
        settings = AppSettingsEntity(api, APP_ID, data)
        # Act
        await settings.refresh()
        # Assert
        assert settings.settings == {'test': 'test'}

    @staticmethod
    @pytest.mark.asyncio
    async def test_refresh_no_app_id(api):
        """Tests refresh when there's no app id."""
        # Arrange
        settings = AppSettingsEntity(api, None)
        # Act/Assert
        with pytest.raises(ValueError):
            await settings.refresh()

    @staticmethod
    @pytest.mark.asyncio
    async def test_save(api):
        """Tests the save function."""
        # Arrange
        data = {'settings': {'test': 'test'}}
        settings = AppSettingsEntity(api, APP_ID, data)
        # Act
        await settings.save()
        # Assert
        assert settings.settings == {'test': 'test'}

    @staticmethod
    @pytest.mark.asyncio
    async def test_save_no_app_id(api):
        """Tests save when there's no app id."""
        # Arrange
        settings = AppSettingsEntity(api, None)
        # Act/Assert
        with pytest.raises(ValueError):
            await settings.save()


class TestAppEntity:
    """Tests for the AppEntity class."""

    @staticmethod
    @pytest.mark.asyncio
    async def test_refresh(api):
        """Tests data is refreshed."""
        # Arrange
        data = get_json('apps.json')['items'][0]
        app = AppEntity(api, data)
        # Act
        await app.refresh()
        # Assert
        assert app.single_instance
        assert app.webhook_target_url == \
            "https://homeassistant.sayre.net:8321/"
        assert app.webhook_public_key

    @staticmethod
    @pytest.mark.asyncio
    async def test_save(api):
        """Tests updating an entity."""
        # Arrange
        data = get_json('app_get.json')
        app = AppEntity(api, data)
        before = app.last_updated_date
        # Act
        await app.save()
        # Assert
        assert app.last_updated_date > before

    @staticmethod
    @pytest.mark.asyncio
    async def test_oauth(api):
        """Tests the oauth method."""
        # Arrange
        data = get_json('app_get.json')
        app = AppEntity(api, data)
        # Act
        oauth = await app.oauth()
        # Assert
        assert oauth.app_id == app.app_id
        assert oauth.client_name == 'pysmartthings-test'
        assert oauth.scope == ["r:devices"]

    @staticmethod
    @pytest.mark.asyncio
    async def test_settings(api):
        """Tests the settings method."""
        # Arrange
        data = get_json('app_get.json')
        app = AppEntity(api, data)
        # Act
        settings = await app.settings()
        # Assert
        assert settings.settings == {'test': 'test'}


class TestOAuth:
    """Tests for the OAuth class."""

    @staticmethod
    def test_init():
        """Tests the initialization."""
        # Arrange
        app_id = '5c03e518-118a-44cb-85ad-7877d0b302e4'
        # Act
        oauth = AppOAuth(app_id)
        # Assert
        assert app_id == oauth.app_id
        assert oauth.scope is not None

    @staticmethod
    def test_client_name():
        """Tests get/set of client name."""
        # Arrange
        oauth = AppOAuth('5c03e518-118a-44cb-85ad-7877d0b302e4')
        # Act
        expected = "my_app"
        oauth.client_name = expected
        actual = oauth.client_name
        # Assert
        assert actual == expected

    @staticmethod
    def test_client_name_invalid():
        """Tests setting an invalid client name."""
        # Arrange
        oauth = AppOAuth('5c03e518-118a-44cb-85ad-7877d0b302e4')
        # Act/Assert
        with pytest.raises(ValueError):
            oauth.client_name = ''


class TestOAuthEntity:
    """Tests for the OAuthEntity class."""

    @staticmethod
    @pytest.mark.asyncio
    async def test_refresh(api):
        """Tests the refresh method."""
        # Arrange
        entity = AppOAuthEntity(
            api, 'c6cde2b0-203e-44cf-a510-3b3ed4706996', None)
        # Act
        await entity.refresh()
        # Assert
        assert entity.client_name == 'pysmartthings-test'
        assert 'r:devices' in entity.scope

    @staticmethod
    @pytest.mark.asyncio
    async def test_save(api):
        """Tests the refresh method."""
        # Arrange
        entity = AppOAuthEntity(
            api, 'c6cde2b0-203e-44cf-a510-3b3ed4706996', None)
        entity.client_name = 'pysmartthings-test'
        entity.scope.append('r:devices')
        # Act/Assert
        await entity.save()



"""Define errors that can be returned from the SmartThings API."""
import json
from typing import Optional, Sequence

from aiohttp import ClientResponseError

UNAUTHORIZED_ERROR = \
    "Authorization for the API is required, but the request has not been " \
    "authenticated."
FORBIDDEN_ERROR = \
    "The request has been authenticated but does not have appropriate" \
    "permissions, or a requested resource is not found."
UNKNOWN_ERROR = "An unknown API error occurred."


class APIErrorDetail:
    """Define details about an error."""

    def __init__(self, data):
        """Create a new instance of the error detail."""
        self._code = data.get('code')
        self._message = data.get('message')
        self._target = data.get('target')
        self._details = []
        details = data.get('details')
        if isinstance(details, list):
            self._details.extend(
                [APIErrorDetail(detail) for detail in details])

    @property
    def code(self) -> Optional[str]:
        """Get the SmartThings-defined error code."""
        return self._code

    @property
    def message(self) -> Optional[str]:
        """Get a description of the error."""
        return self._message

    @property
    def target(self) -> Optional[str]:
        """Get the target of the particular error."""
        return self._target

    @property
    def details(self) -> Sequence:
        """Get an array of errors that represent related errors."""
        return self._details


class APIResponseError(ClientResponseError):
    """Define an error from the API."""

    def __init__(self, request_info, history, *, status=None, message='',
                 headers=None, data=None):
        """Create a new instance of the API Error."""
        super().__init__(request_info, history, status=status,
                         message=message, headers=headers)
        self._raw_error_response = data
        self._request_id = data.get('requestId')
        self._error = APIErrorDetail(data.get('error', {}))

    def __str__(self):
        """Return a string represenation of the error."""
        return "{} ({}): {}".format(
            self.message, self.status, json.dumps(self._raw_error_response))

    @property
    def raw_error_response(self):
        """Get the raw error response returned."""
        return self._raw_error_response

    @property
    def request_id(self) -> Optional[str]:
        """Get request correlation id."""
        return self._request_id

    @property
    def error(self) -> APIErrorDetail:
        """Get the API error document."""
        return self._error

    def is_target_error(self):
        """Determine if the error is due to an issue with the target."""
        return self.error.code == 'ConstraintViolationError' \
            and len(self.error.details) == 1 \
            and self.error.details[0].code \
            and self.error.details[0].code.startswith('Target')


class APIInvalidGrant(Exception):
    """Define an invalid grant error."""

    pass



"""Tests for the scene module."""
import pytest

from pysmartthings.scene import Scene, SceneEntity

from .utilities import get_json


class TestScene:
    """Tests for the scene class."""

    @staticmethod
    def test_apply_data():
        """Test the init method."""
        # Arrange
        data = get_json('scenes.json')
        scene = Scene()
        scene.apply_data(data['items'][0])
        # Assert
        assert scene.scene_id == '3a570170-7c10-4e5a-bef8-0d02175798f2'
        assert scene.color == '#F7F9FF'
        assert scene.name == 'Test'
        assert scene.location_id == '3b44ae84-a735-4fdd-8edd-fc295f4e1563'
        assert scene.icon == 'st.scenes.wand'


class TestSceneEntity:
    """Tests for the scene entity class."""

    @staticmethod
    @pytest.mark.asyncio
    async def test_execute(api):
        """Tests the execute method."""
        # Arrange
        data = get_json('scenes.json')
        entity = SceneEntity(api, data['items'][1])
        # Act
        result = await entity.execute()
        # Assert
        assert result

    @staticmethod
    @pytest.mark.asyncio
    async def test_refresh(api):
        """Tests the refresh method."""
        entity = SceneEntity(api)
        with pytest.raises(NotImplementedError):
            await entity.refresh()

    @staticmethod
    @pytest.mark.asyncio
    async def test_save(api):
        """Tests the refresh method."""
        entity = SceneEntity(api)
        with pytest.raises(NotImplementedError):
            await entity.save()



"""Tests for the Device file."""

import pytest

from pysmartthings.capability import Attribute, Capability
from pysmartthings.device import (
    DEVICE_TYPE_DTH, DEVICE_TYPE_UNKNOWN, Device, DeviceEntity, DeviceStatus,
    Status)

from .conftest import DEVICE_ID, LOCATION_ID, ROOM_ID
from .utilities import get_json


class TestDevice:
    """Tests for the Device class."""

    @staticmethod
    def test_init():
        """Tests whether the Device class initializes correctly."""
        # Arrange/Act
        device = Device()
        # Assert
        assert device.type == DEVICE_TYPE_UNKNOWN
        assert device.capabilities == []
        assert device.components == {}

    @staticmethod
    def test_apply_data():
        """Tests the apply data method."""
        # Arrange
        data = get_json('device.json')
        device = Device()
        # Act
        device.apply_data(data)
        # Assert
        assert device.device_id == DEVICE_ID
        assert device.name == 'GE In-Wall Smart Dimmer'
        assert device.label == 'Front Porch Lights'
        assert device.location_id == LOCATION_ID
        assert device.room_id == ROOM_ID
        assert device.type == DEVICE_TYPE_DTH
        assert device.device_type_id == '8a9d4b1e3b9b1fe3013b9b206a7f000d'
        assert device.device_type_name == 'Dimmer Switch'
        assert device.device_type_network == 'ZWAVE'
        assert device.capabilities == [
            'switch', 'switchLevel', 'refresh', 'indicator', 'sensor',
            'actuator', 'healthCheck', 'light']
        assert device.components == {
            "bottomButton": ['button'],
            "topButton": ['button']
        }

    @staticmethod
    def test_get_capability():
        """Test the capability retrieval method."""
        # Arrange
        data = get_json('device.json')
        device = Device()
        # Act
        device.apply_data(data)
        # Assert
        assert device.get_capability('switch', 'light') == 'switch'
        assert device.get_capability('foo', 'switch') == 'switch'
        assert device.get_capability('foo', 'bar') is None


class TestDeviceEntity:
    """Tests for the DeviceEntity class."""

    @staticmethod
    @pytest.mark.asyncio
    async def test_refresh(api):
        """Tests the refresh method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        await device.refresh()
        # Assert
        assert device.label == 'Front Porch Lights'

    @staticmethod
    @pytest.mark.asyncio
    async def test_save(api):
        """Tests the save method."""
        # Arrange
        device = DeviceEntity(api)
        # Act/Assert
        with pytest.raises(NotImplementedError):
            await device.save()

    @staticmethod
    @pytest.mark.asyncio
    async def test_switch_on(api):
        """Tests the switch_on method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.switch_on()
        # Assert
        assert result
        assert not device.status.switch

    @staticmethod
    @pytest.mark.asyncio
    async def test_switch_on_update(api):
        """Tests the switch_on method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.switch_on(True)
        # Assert
        assert result
        assert device.status.switch

    @staticmethod
    @pytest.mark.asyncio
    async def test_switch_off(api):
        """Tests the switch_on method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.status.switch = True
        # Act
        result = await device.switch_off()
        # Assert
        assert result
        assert device.status.switch

    @staticmethod
    @pytest.mark.asyncio
    async def test_switch_off_update(api):
        """Tests the switch_on method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.status.switch = True
        # Act
        result = await device.switch_off(True)
        # Assert
        assert result
        assert not device.status.switch

    @staticmethod
    @pytest.mark.asyncio
    async def test_lock(api):
        """Tests the lock method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.lock()
        # Assert
        assert result
        assert not device.status.lock

    @staticmethod
    @pytest.mark.asyncio
    async def test_lock_update(api):
        """Tests the lock method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.lock(True)
        # Assert
        assert result
        assert device.status.lock == 'locked'

    @staticmethod
    @pytest.mark.asyncio
    async def test_unlock(api):
        """Tests the unlock method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.unlock()
        # Assert
        assert result
        assert not device.status.lock

    @staticmethod
    @pytest.mark.asyncio
    async def test_unlock_update(api):
        """Tests the unlock method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.unlock(True)
        # Assert
        assert result
        assert device.status.lock == 'unlocked'

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_level(api):
        """Tests the set_level method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_level(75, 2)
        # Assert
        assert result
        assert device.status.level == 0
        assert not device.status.switch

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_level_invalid(api):
        """Tests the set_level method invalid values."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Assert level
        levels = [-1, 101]
        for level in levels:
            with pytest.raises(ValueError):
                await device.set_level(level)
        # Assert duration
        with pytest.raises(ValueError):
            await device.set_level(100, -1)

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_level_update(api):
        """Tests the set_level method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_level(75, 2, True)
        # Assert
        assert result
        assert device.status.level == 75
        assert device.status.switch

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_fan_speed(api):
        """Tests the set_fan_speed method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_fan_speed(66)
        # Assert
        assert result
        assert device.status.level == 0
        assert not device.status.switch

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_fan_speed_invalid(api):
        """Tests the set_fan_speed method invalid values."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Assert
        with pytest.raises(ValueError):
            await device.set_fan_speed(-1)

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_fan_speed_update(api):
        """Tests the set_fan_speed method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_fan_speed(66, True)
        # Assert
        assert result
        assert device.status.fan_speed == 66
        assert device.status.switch

    @staticmethod
    def test_status():
        """Tests the set_level method."""
        # Arrange
        device = DeviceEntity(None, device_id=DEVICE_ID)
        # Act
        status = device.status
        # Assert
        assert status.device_id == DEVICE_ID

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_color_temperature(api):
        """Tests the set_color_temperature method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_color_temperature(3000)
        # Assert
        assert result
        assert device.status.color_temperature == 1

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_color_temperature_invalid(api):
        """Tests the set_color_temperature method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        values = [0, 30001]
        for value in values:
            with pytest.raises(ValueError):
                await device.set_color_temperature(value)

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_color_temperature_update(api):
        """Tests the set_color_temperature method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_color_temperature(3000, True)
        # Assert
        assert result
        assert device.status.color_temperature == 3000

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_hue(api):
        """Tests the set_hue method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_hue(75)
        # Assert
        assert result
        assert device.status.hue == 0

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_hue_invalid(api):
        """Tests the set_hue method invalid values."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Assert
        levels = [-1, 101]
        for level in levels:
            with pytest.raises(ValueError):
                await device.set_hue(level)

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_hue_update(api):
        """Tests the set_hue method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_hue(75, True)
        # Assert
        assert result
        assert device.status.hue == 75

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_saturation(api):
        """Tests the set_saturation method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_saturation(75)
        # Assert
        assert result
        assert device.status.saturation == 0

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_saturation_invalid(api):
        """Tests the set_saturation method invalid values."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Assert
        levels = [-1, 101]
        for level in levels:
            with pytest.raises(ValueError):
                await device.set_saturation(level)

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_saturation_update(api):
        """Tests the set_saturation method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_saturation(75, True)
        # Assert
        assert result
        assert device.status.saturation == 75

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_color(api):
        """Tests the set_color method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_color(25, 50)
        # Assert
        assert result
        assert device.status.hue == 0
        assert device.status.saturation == 0
        assert device.status.color is None

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_color_invalid(api):
        """Tests the set_saturation method invalid values."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Assert
        values = [-1, 101]
        for value in values:
            with pytest.raises(ValueError):
                await device.set_color(value, 0)
            with pytest.raises(ValueError):
                await device.set_color(0, value)

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_color_update(api):
        """Tests the set_saturation method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_color(25, 50, set_status=True)
        # Assert
        assert result
        assert device.status.hue == 25
        assert device.status.saturation == 50
        assert device.status.color == '#4B6432'

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_color_hex(api):
        """Tests the set_color method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_color(color_hex='#4B6432')
        # Assert
        assert result
        assert device.status.hue == 0
        assert device.status.saturation == 0
        assert device.status.color is None

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_color_hex_invalid(api):
        """Tests the set_color method invalid values."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Assert
        values = ['000000', '#00000', '#G00000']
        for value in values:
            with pytest.raises(ValueError):
                await device.set_color(color_hex=value)

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_color_update_hex(api):
        """Tests the set_saturation method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_color(color_hex='#4B6432', set_status=True)
        # Assert
        assert result
        assert device.status.hue == 25
        assert device.status.saturation == 50
        assert device.status.color == '#4B6432'

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_thermostat_fan_mode_legacy(api):
        """Tests the set_saturation method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat)
        device.status.thermostat_fan_mode = 'on'
        # Act
        result = await device.set_thermostat_fan_mode('auto')
        # Assert
        assert result
        assert device.status.thermostat_fan_mode != 'auto'

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_thermostat_fan_mode(api):
        """Tests the set_saturation method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat_fan_mode)
        device.status.thermostat_fan_mode = 'on'
        # Act
        result = await device.set_thermostat_fan_mode('auto')
        # Assert
        assert result
        assert device.status.thermostat_fan_mode != 'auto'

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_thermostat_fan_mode_update(api):
        """Tests the set_saturation method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat_fan_mode)
        device.status.thermostat_fan_mode = 'on'
        # Act
        result = await device.set_thermostat_fan_mode('auto', set_status=True)
        # Assert
        assert result
        assert device.status.thermostat_fan_mode == 'auto'

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_thermostat_mode_legacy(api):
        """Tests the set_thermostat_mode method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat)
        device.status.thermostat_mode = 'heat'
        # Act
        result = await device.set_thermostat_mode('auto')
        # Assert
        assert result
        assert device.status.thermostat_mode != 'auto'

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_thermostat_mode(api):
        """Tests the set_thermostat_mode method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat_mode)
        device.status.thermostat_mode = 'heat'
        # Act
        result = await device.set_thermostat_mode('auto')
        # Assert
        assert result
        assert device.status.thermostat_mode != 'auto'

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_thermostat_mode_update(api):
        """Tests the set_thermostat_mode method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat_mode)
        device.status.thermostat_mode = 'heat'
        # Act
        result = await device.set_thermostat_mode('auto', set_status=True)
        # Assert
        assert result
        assert device.status.thermostat_mode == 'auto'

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_cooling_setpoint_legacy(api):
        """Tests the set_cooling_setpoint method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat)
        device.status.cooling_setpoint = 70
        # Act
        result = await device.set_cooling_setpoint(76)
        # Assert
        assert result
        assert device.status.cooling_setpoint != 76

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_cooling_setpoint_mode(api):
        """Tests the set_cooling_setpoint method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat_cooling_setpoint)
        device.status.cooling_setpoint = 70
        # Act
        result = await device.set_cooling_setpoint(76)
        # Assert
        assert result
        assert device.status.cooling_setpoint != 76

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_cooling_setpoint_update(api):
        """Tests the set_cooling_setpoint method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat_cooling_setpoint)
        device.status.cooling_setpoint = 70
        # Act
        result = await device.set_cooling_setpoint(76, set_status=True)
        # Assert
        assert result
        assert device.status.cooling_setpoint == 76

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_heating_setpoint_legacy(api):
        """Tests the set_heating_setpoint method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat)
        device.status.heating_setpoint = 70
        # Act
        result = await device.set_heating_setpoint(76)
        # Assert
        assert result
        assert device.status.heating_setpoint != 76

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_heating_setpoint_mode(api):
        """Tests the set_heating_setpoint method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat_heating_setpoint)
        device.status.heating_setpoint = 70
        # Act
        result = await device.set_heating_setpoint(76)
        # Assert
        assert result
        assert device.status.heating_setpoint != 76

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_heating_setpoint_update(api):
        """Tests the set_heating_setpoint method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat_heating_setpoint)
        device.status.heating_setpoint = 70
        # Act
        result = await device.set_heating_setpoint(76, set_status=True)
        # Assert
        assert result
        assert device.status.heating_setpoint == 76

    @staticmethod
    @pytest.mark.asyncio
    async def test_open(api):
        """Tests the open method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.door_control)
        # Act/Assert
        assert await device.open()
        assert device.status.door is None
        assert await device.open(set_status=True)
        assert device.status.door == 'opening'

    @staticmethod
    @pytest.mark.asyncio
    async def test_open_legacy(api):
        """Tests the open method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.garage_door_control)
        # Act/Assert
        assert await device.open()
        assert device.status.door is None
        assert await device.open(set_status=True)
        assert device.status.door == 'opening'

    @staticmethod
    @pytest.mark.asyncio
    async def test_close(api):
        """Tests the close method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.door_control)
        # Act/Assert
        assert await device.close()
        assert device.status.door is None
        assert await device.close(set_status=True)
        assert device.status.door == 'closing'

    @staticmethod
    @pytest.mark.asyncio
    async def test_close_legacy(api):
        """Tests the close method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.garage_door_control)
        # Act/Assert
        assert await device.close()
        assert device.status.door is None
        assert await device.close(set_status=True)
        assert device.status.door == 'closing'

    @staticmethod
    @pytest.mark.asyncio
    async def test_open_window_shade(api):
        """Tests the open method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.window_shade)
        # Act/Assert
        assert await device.open()
        assert device.status.window_shade is None
        assert await device.open(set_status=True)
        assert device.status.window_shade == 'opening'

    @staticmethod
    @pytest.mark.asyncio
    async def test_close_window_shade(api):
        """Tests the close method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.window_shade)
        # Act/Assert
        assert await device.close()
        assert device.status.window_shade is None
        assert await device.close(set_status=True)
        assert device.status.window_shade == 'closing'

    @staticmethod
    @pytest.mark.asyncio
    async def test_request_drlc_action(api):
        """Tests the request_drlc_action method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act/Assert
        assert await device.request_drlc_action(
            1, 2, '1970-01-01T00:00:00Z', 10, 1)
        assert device.status.drlc_status is None
        assert await device.request_drlc_action(
            1, 2, '1970-01-01T00:00:00Z', 10, 1, set_status=True)
        assert device.status.drlc_status == {
            "duration": 10,
            "drlcLevel": 2,
            "start": '1970-01-01T00:00:00Z',
            "override": False
        }

    @staticmethod
    @pytest.mark.asyncio
    async def test_override_drlc_action(api):
        """Tests the override_drlc_action method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act/Assert
        assert await device.override_drlc_action(True)
        assert device.status.drlc_status is None
        assert await device.override_drlc_action(True, set_status=True)
        assert device.status.drlc_status == {
            "override": True
        }

    @staticmethod
    @pytest.mark.asyncio
    async def test_execute(api):
        """Tests the execute method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act/Assert
        assert await device.execute('Test', {'data': 'Test'})

    @staticmethod
    @pytest.mark.asyncio
    async def test_preset_position(api):
        """Tests the preset_position method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act/Assert
        assert await device.preset_position()

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_air_conditioner_mode(api):
        """Tests the set_air_conditioner_mode method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act/Assert
        assert await device.set_air_conditioner_mode('auto')
        assert device.status.air_conditioner_mode is None
        assert await device.set_air_conditioner_mode('auto', set_status=True)
        assert device.status.air_conditioner_mode == 'auto'

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_fan_mode(api):
        """Tests the set_fan_mode method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act/Assert
        assert await device.set_fan_mode('auto')
        assert device.status.fan_mode is None
        assert await device.set_fan_mode('auto', set_status=True)
        assert device.status.fan_mode == 'auto'

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_air_flow_direction(api):
        """Tests the set_air_flow_direction method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act/Assert
        assert await device.set_air_flow_direction('fixed')
        assert device.status.air_flow_direction is None
        assert await device.set_air_flow_direction('fixed', set_status=True)
        assert device.status.air_flow_direction == 'fixed'


class TestDeviceStatus:
    """Tests for the DeviceStatus class."""

    @staticmethod
    def test_init():
        """Tests the init method."""
        # Arrange/Act
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # Assert
        assert status.device_id == DEVICE_ID
        assert status.attributes == {}
        assert not status.switch
        assert not status.motion
        assert status.level == 0
        assert status.component_id == 'main'

    @staticmethod
    def test_apply_data():
        """Tests the apply_data method."""
        # Arrange
        data = get_json('device_status.json')
        # Act
        status = DeviceStatus(None, DEVICE_ID, data)
        # Assert
        assert len(status.attributes) == 9
        assert status.switch
        assert status.level == 100
        assert len(status.components) == 2
        assert len(status.components['topButton'].attributes) == 3
        assert len(status.components['bottomButton'].attributes) == 3

    @staticmethod
    def test_apply_attribute_update():
        """Tests the apply_attribute_update method."""
        # Arrange
        data = get_json('device_status.json')
        device = DeviceStatus(None, DEVICE_ID, data)
        # Act
        device.apply_attribute_update(
            'main', Capability.switch_level, Attribute.level, 50, '%',
            {'test': 'test'})
        # Assert
        status = device.attributes[Attribute.level]
        assert status.value == 50
        assert status.unit == '%'
        assert status.data == {'test': 'test'}

    @staticmethod
    def test_apply_attribute_update_preserve_unit():
        """Tests the apply_attribute_update preserves the old unit."""
        # Arrange
        data = get_json('device_status.json')
        device = DeviceStatus(None, DEVICE_ID, data)
        device.attributes[Capability.switch_level] = Status(40, '%', None)
        # Act
        device.apply_attribute_update(
            'main', Capability.switch_level, Attribute.level, 50)
        # Assert
        status = device.attributes[Attribute.level]
        assert status.unit == '%'

    @staticmethod
    def test_apply_attribute_update_child_status():
        """Tests the apply_attribute_update method to a child status."""
        # Arrange
        data = get_json('device_status.json')
        status = DeviceStatus(None, DEVICE_ID, data)
        # Act
        status.apply_attribute_update(
            'bottomButton', 'switchLevel', 'level', 50)
        # Assert
        assert status.components['bottomButton'].level == 50

    @staticmethod
    def test_values():
        """Test the values property."""
        # Arrange
        data = get_json('device_status.json')
        status = DeviceStatus(None, DEVICE_ID, data)
        # Act/Assert
        assert status.values == {
            'button': None,
            'numberOfButtons': None,
            'supportedButtonValues': None,
            'indicatorStatus': 'when off',
            'switch': 'on',
            'checkInterval': 1920,
            'healthStatus': None,
            'DeviceWatch-DeviceStatus': None,
            'level': 100
        }

    @staticmethod
    def test_attributes():
        """Test the attributes property."""
        # Arrange
        data = get_json('device_status.json')
        status = DeviceStatus(None, DEVICE_ID, data)
        # Act/Assert
        assert status.attributes == {
            'button': (None, None, None),
            'numberOfButtons': (None, None, None),
            'supportedButtonValues': (None, None, None),
            'indicatorStatus': ('when off', None, None),
            'switch': ('on', None, None),
            'checkInterval': (1920, 's', {"protocol": "zwave",
                                          "hubHardwareId": "000F"}),
            'healthStatus': (None, None, {}),
            'DeviceWatch-DeviceStatus': (None, None, {}),
            'level': (100, '%', None)
        }

    @staticmethod
    def test_attributes_default():
        """Test the attributes property."""
        # Arrange
        data = get_json('device_status.json')
        status = DeviceStatus(None, DEVICE_ID, data)
        # Act/Assert
        assert status.attributes['thermostatSetpoint'] == (None, None, None)

    @staticmethod
    @pytest.mark.asyncio
    async def test_refresh(api):
        """Tests the refresh method."""
        # Arrange
        status = DeviceStatus(api, device_id=DEVICE_ID)
        # Act
        await status.refresh()
        # Assert
        assert len(status.attributes) == 9

    @staticmethod
    def test_switch():
        """Tests the switch property."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # Act
        status.switch = True
        # Assert
        assert status.switch

    @staticmethod
    def test_level():
        """Tests the level property."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # Act
        status.level = 50
        # Assert
        assert status.level == 50

    @staticmethod
    def test_level_range():
        """Tests the level property's range."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # Act/Assert
        values = [-1, 101]
        for value in values:
            with pytest.raises(ValueError):
                status.level = value

    @staticmethod
    def test_fan_speed():
        """Tests the fan_speed property."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # Act
        status.fan_speed = 50
        # Assert
        assert status.fan_speed == 50

    @staticmethod
    def test_fan_speed_range():
        """Tests the fan_speed property's range."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # Act/Assert
        with pytest.raises(ValueError):
            status.fan_speed = -1

    @staticmethod
    def test_hue_range():
        """Tests the hue property's range."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # Act/Assert
        values = [-1, 101]
        for value in values:
            with pytest.raises(ValueError):
                status.hue = value

    @staticmethod
    def test_saturation_range():
        """Tests the hue property's range."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # Act/Assert
        values = [-1, 101]
        for value in values:
            with pytest.raises(ValueError):
                status.saturation = value

    @staticmethod
    def test_color_temperature_range():
        """Tests the hue property's range."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # Act/Assert
        values = [0, 30001]
        for value in values:
            with pytest.raises(ValueError):
                status.color_temperature = value

    @staticmethod
    def test_color_format():
        """Tests the color property's validation."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # Act/Assert
        values = ['000000', '#00000', '#HH2000']
        for value in values:
            with pytest.raises(ValueError):
                status.color = value

    @staticmethod
    def test_is_on():
        """Tests the is_on method."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)
        status.update_attribute_value(Attribute.acceleration, 'active')
        status.level = 100
        # Act/Assert
        assert status.is_on(Attribute.acceleration)
        assert status.is_on(Attribute.level)
        assert not status.is_on(Attribute.switch)

    @staticmethod
    def test_well_known_attributes():
        """Tests the humidity property."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)

        assert status.supported_ac_modes == []
        assert status.supported_ac_fan_modes == []

        status.update_attribute_value(Attribute.humidity, 50)
        status.update_attribute_value(Attribute.temperature, 55)
        status.update_attribute_value(
            Attribute.thermostat_operating_state, 'on')
        status.update_attribute_value(
            Attribute.supported_thermostat_fan_modes, ['on', 'off'])
        status.update_attribute_value(
            Attribute.supported_thermostat_modes, ['auto', 'off'])
        status.update_attribute_value(Attribute.lock, 'locked')
        status.update_attribute_value(Attribute.door, 'open')
        status.update_attribute_value(Attribute.window_shade, 'closed')
        status.update_attribute_value(Attribute.data, {'test': 'test'})
        status.update_attribute_value(Attribute.three_axis, [0, 0, 0])
        status.update_attribute_value(
            Attribute.supported_ac_modes, ['auto', 'cool'])
        status.update_attribute_value(Attribute.fan_mode, 'low')
        status.update_attribute_value(
            Attribute.supported_ac_fan_modes, ['auto', 'low'])
        # Act/Assert
        assert status.humidity == 50
        assert status.temperature == 55
        assert status.thermostat_operating_state == 'on'
        assert status.supported_thermostat_fan_modes == ['on', 'off']
        assert status.supported_thermostat_modes == ['auto', 'off']
        assert status.lock == 'locked'
        assert status.door == 'open'
        assert status.window_shade == 'closed'
        assert status.data == {'test': 'test'}
        assert status.three_axis == [0, 0, 0]
        assert status.supported_ac_modes == ['auto', 'cool']
        assert status.supported_ac_fan_modes == ['auto', 'low']

    @staticmethod
    def test_well_known_ocf_attributes():
        """Tests the OCF related attributes."""
        # Arrange
        data = get_json('device_samsungac_status.json')
        status = DeviceStatus(None, device_id=DEVICE_ID, data=data)
        # Act/Assert
        assert status.ocf_data_model_version == 'res.1.1.0,sh.1.1.0'
        assert status.ocf_date_of_manufacture == '2019-02-26T02:05:55Z'
        assert status.ocf_device_id == DEVICE_ID
        assert status.ocf_firmware_version == '0.1.0'
        assert status.ocf_hardware_version == '1.0'
        assert status.ocf_manufacturer_details_link == 'http://www.samsung.com'
        assert status.ocf_manufacturer_name == 'Samsung Electronics'
        assert status.ocf_model_number == \
            'ARTIK051_KRAC_18K|10193441|60010123001111010200000000000000'
        assert status.ocf_name == 'Air Conditioner'
        assert status.ocf_os_version == 'TizenRT2.0'
        assert status.ocf_platform_id == 'd5226d90-1b4f-e59d-5f3f-027ac3b18faf'
        assert status.ocf_platform_version == '0.1.0'
        assert status.ocf_spec_version == 'core.1.1.0'
        assert status.ocf_support_link == 'http://www.samsung.com/support'
        assert status.ocf_system_time == '02:05:55Z'
        assert status.ocf_vendor_id == 'DA-AC-RAC-000001'

    @staticmethod
    def test_well_known_drlc_attributes():
        """Tests the drlc related attributes."""
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # No attribute
        assert status.drlc_status is None
        assert status.drlc_status_duration is None
        assert status.drlc_status_level is None
        assert status.drlc_status_override is None
        assert status.drlc_status_start is None
        # Populated
        drlc_status = {
            "duration": 0,
            "drlcLevel": -1,
            "start": "1970-01-01T00:00:00Z",
            "override": False
        }
        status.update_attribute_value(Attribute.drlc_status, drlc_status)
        assert status.drlc_status == drlc_status
        assert status.drlc_status_duration == 0
        assert status.drlc_status_level == -1
        assert not status.drlc_status_override
        assert status.drlc_status_start == '1970-01-01T00:00:00Z'
        # Missing
        status.update_attribute_value(Attribute.drlc_status, {})
        assert status.drlc_status == {}
        assert status.drlc_status_duration is None
        assert status.drlc_status_level is None
        assert status.drlc_status_override is None
        assert status.drlc_status_start is None
        # Not valid
        drlc_status = {
            "duration": 'Foo',
            "drlcLevel": 'Foo'
        }
        status.update_attribute_value(Attribute.drlc_status, drlc_status)
        assert status.drlc_status == drlc_status
        assert status.drlc_status_duration is None
        assert status.drlc_status_level is None

    @staticmethod
    def test_well_known_power_consumption_attributes():
        """Tests the power consumption related attributes."""
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # No attribute
        assert status.power_consumption is None
        assert status.power_consumption_end is None
        assert status.power_consumption_energy is None
        assert status.power_consumption_power is None
        assert status.power_consumption_start is None
        # Populated
        data = {
            "start": "2019-02-24T21:03:04Z",
            "power": 0,
            "energy": 500,
            "end": "2019-02-26T02:05:55Z"
        }
        status.update_attribute_value(Attribute.power_consumption, data)
        assert status.power_consumption == data
        assert status.power_consumption_end == '2019-02-26T02:05:55Z'
        assert status.power_consumption_energy == 500
        assert status.power_consumption_power == 0
        assert status.power_consumption_start == '2019-02-24T21:03:04Z'
        # Missing
        status.update_attribute_value(Attribute.power_consumption, {})
        assert status.power_consumption == {}
        assert status.power_consumption_end is None
        assert status.power_consumption_energy is None
        assert status.power_consumption_power is None
        assert status.power_consumption_start is None
        # Not valid
        data = {
            "power": "Foo",
            "energy": "Bar"
        }
        status.update_attribute_value(Attribute.power_consumption, data)
        assert status.power_consumption == data
        assert status.power_consumption_energy is None
        assert status.power_consumption_power is None



"""SmartThings Cloud API"""

import os

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join("README.md"), 'r') as fh:
    long_description = fh.read()

const = {}
with open(os.path.join('pysmartthings', 'const.py'), 'r') as fp:
    exec(fp.read(), const)

setup(name=const['__title__'],
      version=const['__version__'],
      description='A python library for interacting with the SmartThings cloud API build with asyncio and aiohttp.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/andrewsayre/pysmartthings',
      author='Andrew Sayre',
      author_email='andrew@sayre.net',
      license='ASL 2.0',
      packages=find_packages(),
      install_requires=['aiohttp>=3.5.1,<4.0.0'],
      tests_require=['tox>=3.5.0,<4.0.0'],
      platforms=['any'],
      keywords="smartthings",
      zip_safe=False,
      classifiers=[
          "Development Status :: 4 - Beta",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: Apache Software License",
          "Operating System :: OS Independent",
          "Topic :: Software Development :: Libraries",
          "Topic :: Home Automation",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          ])
